"""LoRaWAN module

This module provides a set of tools upon the basic LoRa PHY:

- An emulated single-channel LoRaWAN gateway implementation that can run a LoRaWAN application ;
- A modular LoRaWAN channel plan class (ChannelPlan) that can describe the different uplink and downlink channels as
  well as the supported datarate for each of them ;
- A hackable LoRaWAN gateway stack using WHAD stack model
"""
from whad.phy.connector.lora import LoRa

from .helpers import EUI
from .channel import ChannelPlan, Uplink, Downlink, DataRate
from .connector.gateway import LWGateway
from .app import LWApplication
from .stack import LWGatewayStack
from .exceptions import ChannelNotFound, InvalidDataRate, InvalidNodeRegistryError, \
    BadEuiFormat, BadMICError, NotStartedException, MissingKeyError

__all__ = [
    'LoRa',
    'EUI',
    'ChannelPlan',
    'Uplink',
    'Downlink',
    'DataRate',
    'LWApplication',
    'LWGateway',
    'LWGatewayStack',
    'ChannelNotFound',
    'InvalidDataRate',
    'InvalidNodeRegistryError',
    'BadEuiFormat',
    'BadMICError',
    'NotStartedException',
    'MissingKeyError'
]



from whad.unifying.hid.exceptions import InvalidHIDData
from whad.common.converters.hid import HIDConverter, HIDLocaleNotFound, HIDCodeNotFound
from whad.helpers import bytes_to_bits
from struct import pack

class LogitechUnifyingKeystrokeConverter(HIDConverter):
    @classmethod
    def get_key_from_hid_data(cls, hid_data, locale="fr"):

        if not isinstance(hid_data, bytes):
            raise InvalidHIDData(hid_data)

        if len(hid_data) != 7:
            raise InvalidHIDData(hid_data)

        modifiers = hid_data[0]
        hid_code = hid_data[1]
        return super().get_key_from_hid_code(hid_code, modifiers, locale=locale)

    @classmethod
    def get_hid_data_from_key(cls, key, ctrl=False, alt=False, shift=False, gui=False, locale="fr"):
        (hid_code, modifiers) = super().get_hid_code_from_key(key, ctrl=ctrl, alt=alt, shift=shift, gui=gui, locale=locale)

        hid_data = pack("B", modifiers) + pack("B", hid_code) + b"\x00"*5

        return hid_data

class LogitechUnifyingMouseMovementConverter:
    @classmethod
    def get_coordinates_from_hid_data(cls, hid_data):

        if not isinstance(hid_data, bytes):
            raise InvalidHIDData(hid_data)

        if len(hid_data) != 3:
            raise InvalidHIDData(hid_data)

        bits = bytes_to_bits(hid_data)
        xb = bits[12:16] + bits[0:8]
        yb =  bits[16:] + bits[8:12]

        if xb[0] == "0":
            x = sum([(2**(11-i))*int(xb[i]) for i in range(0,12)])
        else:
            x = -1*(1+sum([(2**(11-i))*(1 - int(xb[i])) for i in range(0,12)]))

        if yb[0] == "0":
            y = sum([(2**(11-i))*int(yb[i]) for i in range(0,12)])
        else:
            y = -1*(1+sum([(2**(11-i))*(1 - int(yb[i])) for i in range(0,12)]))

        return (x, y)

    @classmethod
    def get_hid_data_from_coordinates(cls, x, y):
        if (y < 0):
            y += 4096
        if (x < 0):
            x += 4096

        a,b,c = 0,0,0
        a = x & 0xFF

        b |= (x >> 8) & 0x0F
        c = (y >> 4) & 0xFF
        b |= (y << 4) & 0xF0

        ab = pack('B',a)
        bb = pack('B',b)
        cb = pack('B',c)

        return ab + bb + cb



"""
BLE ATT layer manager

This layer manager mostly translates Scapy ATT packets into GATT messages required by the GATT
layer, and exposes an interface to the GATT layer in order to allow ATT packets to be forged
and sent to the underlying layer (L2CAP).
"""
import logging

from typing import List
from scapy.layers.bluetooth import ATT_Error_Response, ATT_Exchange_MTU_Request, \
    ATT_Exchange_MTU_Response, ATT_Execute_Write_Request, ATT_Execute_Write_Response, \
    ATT_Find_By_Type_Value_Request, ATT_Find_By_Type_Value_Response, ATT_Find_Information_Request, \
    ATT_Find_Information_Response, ATT_Prepare_Write_Request, ATT_Prepare_Write_Response, \
    ATT_Read_Blob_Request, ATT_Handle_Value_Indication, ATT_Handle_Value_Notification, \
    ATT_Read_Blob_Response, ATT_Read_By_Group_Type_Request, ATT_Read_By_Group_Type_Response, \
    ATT_Read_By_Type_Request, ATT_Read_By_Type_Response, ATT_Read_Multiple_Request, \
    ATT_Read_Multiple_Response, ATT_Read_Request, ATT_Read_Response, ATT_Write_Command, \
    ATT_Write_Response, ATT_Write_Request, ATT_Read_By_Type_Request_128bit, ATT_Hdr, \
    ATT_Handle
from scapy.packet import bind_layers, Packet

from whad.ble.stack.att.constants import BleAttOpcode, SecurityProperty, SecurityAccess

from whad.common.stack import Layer, ContextualLayer, alias, instance

#from whad.ble.stack.gatt import  Gatt
from whad.ble.stack.gatt.message import GattExecuteWriteRequest, GattExecuteWriteResponse, \
    GattFindInfoResponse, GattHandleValueIndication, GattHandleValueNotification, \
    GattPrepareWriteRequest, GattPrepareWriteResponse, GattReadByGroupTypeResponse, \
    GattErrorResponse, GattFindByTypeValueRequest, GattFindByTypeValueResponse, \
    GattFindInfoRequest, GattReadByTypeResponse, GattReadRequest, GattReadResponse, \
    GattReadBlobRequest, GattReadBlobResponse, GattReadMultipleRequest, \
    GattReadMultipleResponse, GattWriteCommand, GattWriteRequest, GattWriteResponse, \
    GattReadByGroupTypeRequest, GattReadByTypeRequest, GattReadByTypeRequest128, \
    GattExchangeMtuResponse, GattExchangeMtuRequest
from whad.scapy.layers.bluetooth import ATT_Handle_Value_Confirmation

logger = logging.getLogger(__name__)


@alias('att')
class ATTLayer(Layer):
    """ATT layer implementation.
    """

    def configure(self, options=None):
        """Layer configuration
        """
        # Initialize state
        self.state.client_att_mtu = 23
        self.state.server_att_mtu = 23

    def set_client_mtu(self, mtu: int):
        """Set ATT client MTU
        """
        if mtu >= 23:
            self.state.client_att_mtu = mtu
    
    def get_client_mtu(self) -> int:
        """Retrieve ATT client MTU
        """
        return self.state.client_att_mtu

    def set_server_mtu(self, mtu: int):
        """Set ATT server MTU
        """
        if mtu >= 23:
            self.state.server_att_mtu = mtu

    def get_server_mtu(self) -> int:
        """Retrieve ATT server MTU
        """
        return self.state.server_att_mtu
    
    def notify_client_mtu(self):
        """Notify l2cap that ATT MTU has changed for client
        """
        self.send("l2cap", self.state.client_att_mtu, tag="ATT_MTU")

    def notify_server_mtu(self):
        """Notify l2cap that ATT MTU has changed for server
        """
        self.send("l2cap", self.state.server_att_mtu, tag="ATT_MTU")

    ##########################################
    # Incoming requests and responses
    ##########################################

    @instance('l2cap')
    def on_packet(self, inst: ContextualLayer, att_pkt: Packet):
        """Dispatch ATT packet.

        :param inst: L2CAP instance object
        :type inst: ContextualLayer
        :param att_pkt: Incoming ATT packet to process
        :type att_pkt: Packet
        """
        if ATT_Error_Response in att_pkt:
            self.on_error_response(att_pkt.getlayer(ATT_Error_Response))
        elif ATT_Exchange_MTU_Request in att_pkt:
            self.on_exch_mtu_request(att_pkt.getlayer(ATT_Exchange_MTU_Request))
        elif ATT_Exchange_MTU_Response in att_pkt:
            self.on_exch_mtu_response(att_pkt.getlayer(ATT_Exchange_MTU_Response))

        elif ATT_Find_Information_Request in att_pkt:
            self.on_find_info_request(att_pkt.getlayer(ATT_Find_Information_Request))
        elif ATT_Find_Information_Response in att_pkt:
            self.on_find_info_response(att_pkt.getlayer(ATT_Find_Information_Response))
        elif ATT_Find_By_Type_Value_Request in att_pkt:
            self.on_find_by_type_value_request(att_pkt.getlayer(ATT_Find_By_Type_Value_Request))
        elif ATT_Find_By_Type_Value_Response in att_pkt:
            self.on_find_by_type_value_response(att_pkt.getlayer(ATT_Find_By_Type_Value_Response))
        elif ATT_Read_By_Type_Request in att_pkt:
            self.on_read_by_type_request(att_pkt.getlayer(ATT_Read_By_Type_Request))
        elif ATT_Read_By_Type_Response in att_pkt:
            self.on_read_by_type_response(att_pkt.getlayer(ATT_Read_By_Type_Response))
        elif ATT_Read_By_Group_Type_Request in att_pkt:
            self.on_read_by_group_type_request(att_pkt.getlayer(ATT_Read_By_Group_Type_Request))
        elif ATT_Read_By_Group_Type_Response in att_pkt:
            self.on_read_by_group_type_response(att_pkt.getlayer(ATT_Read_By_Group_Type_Response))
        elif ATT_Read_Request in att_pkt:
            self.on_read_request(att_pkt.getlayer(ATT_Read_Request))
        elif ATT_Read_By_Type_Request_128bit in att_pkt:
            self.on_read_by_type_request_128bit(att_pkt.getlayer(ATT_Read_By_Type_Request_128bit))
        elif ATT_Read_Response in att_pkt:
            self.on_read_response(att_pkt.getlayer(ATT_Read_Response))
        elif ATT_Read_Blob_Request in att_pkt:
            self.on_read_blob_request(att_pkt.getlayer(ATT_Read_Blob_Request))
        elif ATT_Read_Blob_Response in att_pkt:
            self.on_read_blob_response(att_pkt.getlayer(ATT_Read_Blob_Response))
        elif ATT_Read_Multiple_Request in att_pkt:
            self.on_read_multiple_request(att_pkt.getlayer(ATT_Read_Multiple_Request))
        elif ATT_Read_Multiple_Response in att_pkt:
            self.on_read_multiple_response(att_pkt.getlayer(ATT_Read_Multiple_Response))
        elif ATT_Write_Request in att_pkt:
            self.on_write_request(att_pkt.getlayer(ATT_Write_Request))
        elif ATT_Write_Response in att_pkt:
            self.on_write_response(att_pkt.getlayer(ATT_Write_Response))
        elif ATT_Write_Command in att_pkt:
            self.on_write_command(att_pkt.getlayer(ATT_Write_Command))
        elif ATT_Handle_Value_Notification in att_pkt:
            self.on_handle_value_notification(att_pkt.getlayer(ATT_Handle_Value_Notification))
        elif ATT_Handle_Value_Indication in att_pkt:
            self.on_handle_value_indication(att_pkt.getlayer(ATT_Handle_Value_Indication))
        elif ATT_Prepare_Write_Request in att_pkt:
            self.on_prepare_write_request(att_pkt.getlayer(ATT_Prepare_Write_Request))
        elif ATT_Prepare_Write_Response in att_pkt:
            self.on_prepare_write_response(att_pkt.getlayer(ATT_Prepare_Write_Response))
        elif ATT_Execute_Write_Request in att_pkt:
            self.on_execute_write_request(att_pkt.getlayer(ATT_Execute_Write_Request))
        elif ATT_Execute_Write_Response in att_pkt:
            self.on_execute_write_response(att_pkt.getlayer(ATT_Execute_Write_Response))
        # Signed command not supported yet
        # Write Response has no body
        elif att_pkt.opcode == BleAttOpcode.WRITE_RESPONSE:
            self.on_write_response(None)
        # Read Blob Response has no body
        elif att_pkt.opcode == BleAttOpcode.READ_BLOB_RESPONSE:
            self.on_read_blob_response(None)
        # Read Response has no body
        elif att_pkt.opcode == BleAttOpcode.READ_RESPONSE:
            self.on_read_response(None)
        # Execute write request
        elif att_pkt.opcode == BleAttOpcode.EXECUTE_WRITE_RESPONSE:
            self.on_execute_write_response(None)

    def on_error_response(self, error_resp: ATT_Error_Response):
        """Process a generic ATT error response.

        :param error_resp: Error response packet
        :type error_resp: ATT_Error_Response
        """
        # Send a GattErrorResponse message to gatt
        self.send(
            'gatt',
            GattErrorResponse(
                error_resp.request,
                error_resp.handle,
                error_resp.ecode
            ),
            tag='GATT_ERROR_RSP',
        )

    def on_exch_mtu_request(self, mtu_req: ATT_Exchange_MTU_Request):
        """Handle ATT Exchange MTU request, update L2CAP TX MTU and returns
        our MTU.

        :param mtu_req: MTU request
        :type mtu_req: ATT_Exchange_MTU_Request
        """
        logger.debug("[att] got an MTU exchange request (mtu: %d)", mtu_req.mtu)

        # Send back our MTU.
        #self.send_data(ATT_Exchange_MTU_Response(
        #    mtu=mtu_req.mtu
        #))

        self.send("gatt", GattExchangeMtuRequest(mtu=mtu_req.mtu), tag="XCHG_MTU_REQ")

    def on_exch_mtu_response(self, mtu_resp: ATT_Exchange_MTU_Response):
        """Update L2CAP remote MTU based on ATT_Exchange_MTU_Response.

        :param mtu_resp: MTU response
        :type mtu_resp: ATT_Exchange_MTU_Request
        """

        # Forward to GATT
        self.send('gatt', GattExchangeMtuResponse(mtu=mtu_resp.mtu),
                  tag='XCHG_MTU_RESP')


    def on_find_info_request(self, request: ATT_Find_Information_Request):
        """Handle ATT Find Information Request

        :param request: Request
        :type request: ATT_Find_Information_Request
        """
        self.send('gatt', GattFindInfoRequest(
                request.start,
                request.end
            ), tag='FIND_INFO_REQ'
        )

    def on_find_info_response(self, response: ATT_Find_Information_Response):
        """Handle ATT Find Information Response

        :param response: Find information response packet
        :type response: ATT_Find_Information_Response
        """
        handles = b''.join([item.build() for item in response.handles])
        self.send('gatt', GattFindInfoResponse.from_bytes(
                response.format,
                handles
            ), tag='FIND_INFO_RESP',
        )

    def on_find_by_type_value_request(self, request: ATT_Find_By_Type_Value_Request):
        """Handle ATT Find By Type Value request

        :param request: Find by type request
        :type request: ATT_Find_Information_Response
        """
        self.send('gatt', GattFindByTypeValueRequest(
                request.start,
                request.end,
                request.uuid,
                request.data
            )
        )

    def on_find_by_type_value_response(self, response: ATT_Find_By_Type_Value_Response):
        """Handle ATT Find by type value response

        :param response: Find by type value response packet
        :type response: ATT_Find_Information_Response
        """
        handles = b''.join([item.build() for item in response.handles])
        self.send('gatt', GattFindByTypeValueResponse.from_bytes(handles))

    def on_read_by_type_request(self, request: ATT_Read_By_Type_Request):
        """Handle read by type request

        :param request: ReadByType packet
        :type request: ATT_Read_By_Type_Request
        """
        self.send('gatt', GattReadByTypeRequest(
            request.start,
            request.end,
            request.uuid
        ))

    def on_read_by_type_request_128bit(self, request: ATT_Read_By_Type_Request_128bit):
        """Handle ATT Read By Type Request 128-bit UUID

        :param request: ReadByType request with 128-bit UUID
        :type request: ATT_Read_By_Type_Request_128bit
        """
        self.send('gatt',
            request.start,
            request.end,
            request.uuid1,
            request.uuid2
        )

    def on_read_by_type_response(self, response: GattReadByTypeResponse):
        """Handle read by type response

        :param response: ReadByTypeResponse
        :type response: GattReadByTypeResponse
        """
        # Must rebuild handles payload as bytes, since scapy parsed it :(
        handles = b''.join([item.build() for item in response.handles])
        self.send('gatt', GattReadByTypeResponse.from_bytes(
                response.len,
                handles
            )
        )

    def on_read_request(self, request: ATT_Read_Request):
        """Handle ATT Read Request

        :param request: Read request
        :type request: ATT_Read_Request
        """
        self.send('gatt', GattReadRequest(
                request.gatt_handle
            )
        )

    def on_read_response(self, response: ATT_Read_Response):
        """Handle ATT Read Response

        :param response: Read response
        :type response: ATT_Read_Response
        """
        if response is not None:
            self.send('gatt', GattReadResponse(
                    response.value
                )
            )
        else:
            self.send('gatt', GattReadResponse(
                    b''
                )
            )

    def on_read_blob_request(self, request: ATT_Read_Blob_Request):
        """Handle ATT Read Blob Request

        :param request: ReadBlobRequest
        :type request: ATT_Read_Blob_Request
        """
        self.send('gatt', GattReadBlobRequest(
                request.gatt_handle,
                request.offset
            )
        )

    def on_read_blob_response(self, response: ATT_Read_Blob_Response):
        """Handle ATT Read Blob Response

        :param response: ATT response if provided, None otherwise.
        :type response: ATT_Read_Blob_Response
        """
        if response is not None:
            self.send('gatt', GattReadBlobResponse(
                    response.value
                )
            )
        else:
            self.send('gatt', GattReadBlobResponse(
                    None
                )
            )

    def on_read_multiple_request(self, request: ATT_Read_Multiple_Request):
        """Handle ATT Read Multiple Request

        :param request: ReadMultiple request
        :type request: ATT_Read_Multiple_Request
        """
        self.send('gatt',GattReadMultipleRequest(request.handles))

    def on_read_multiple_response(self, response: ATT_Read_Multiple_Response):
        """Handle ATT Read Multiple Response

        :param response: ReadMultiple response
        :type response: ATT_Read_Multiple_Response
        """
        self.send('gatt', GattReadMultipleResponse(response.values))

    def on_read_by_group_type_request(self, request: ATT_Read_By_Group_Type_Request):
        """Handle ATT Read By Group Type Request

        :param request: ReadByGroupType request
        :type request: ATT_Read_By_Group_Type_Request
        """
        self.send('gatt', GattReadByGroupTypeRequest(
                request.start,
                request.end,
                request.uuid
            )
        )

    def on_read_by_group_type_response(self, response: ATT_Read_By_Group_Type_Response):
        """Handle ATT Read By Group Type Response

        :param response: ReadByGroupType response
        :type response: ATT_Read_By_Group_Type_Response
        """
        self.send('gatt', GattReadByGroupTypeResponse.from_bytes(
                response.length,
                response.data
            )
        )

    def on_write_request(self, request: ATT_Write_Request):
        """Handle ATT Write Request

        :param request: Write request
        :type request: ATT_Write_Request
        """
        self.send('gatt', GattWriteRequest(
                request.gatt_handle,
                request.data
            )
        )

    def on_write_response(self, response: ATT_Write_Response):
        """Handle ATT Write Response

        :param response: Write response
        :type response: ATT_Write_Response
        """
        self.send('gatt', GattWriteResponse())

    def on_write_command(self, command: ATT_Write_Command):
        """Handle ATT Write Command

        :param command: Write command
        :type command: ATT_Write_Command
        """
        self.send('gatt', GattWriteCommand(
                command.gatt_handle,
                command.data
            )
        )

    def on_prepare_write_request(self, request:ATT_Prepare_Write_Request):
        """Handle ATT Prepare Write Request

        :param request: PrepareWrite request
        :type request: ATT_Prepare_Write_Request
        """
        self.send('gatt', GattPrepareWriteRequest(
                request.gatt_handle,
                request.offset,
                request.data
            )
        )

    def on_prepare_write_response(self, response: ATT_Prepare_Write_Response):
        """Handle ATT Prepare Write Response

        :param response: PrepareWrite response
        :type response: ATT_Prepare_Write_Response
        """
        self.send('gatt', GattPrepareWriteResponse(
                response.gatt_handle,
                response.offset,
                response.data
            )
        )

    def on_execute_write_request(self, request: ATT_Execute_Write_Request):
        """Handle ATT Execute Write Request

        :param request: ExecuteWrite request
        :type request: ATT_Execute_Write_Request
        """
        self.send('gatt', GattExecuteWriteRequest(
                request.flags
            )
        )

    def on_execute_write_response(self, response: ATT_Execute_Write_Response):
        """Handle ATT Execute Write Response

        :param response: ExecuteWrite response
        :type response: ATT_Execute_Write_Response
        """
        self.send('gatt', GattExecuteWriteResponse())

    def on_handle_value_notification(self, notif: ATT_Handle_Value_Notification):
        """Handle ATT Handle Value Notification

        :param notif: Handle value notification packet
        :type notif: ATT_Handle_Value_Notification
        """
        self.send('gatt', GattHandleValueNotification(
                notif.gatt_handle,
                notif.value
            )
        )

    def on_handle_value_indication(self, notif: ATT_Handle_Value_Indication):
        """Handle ATT Handle Value indication

        :param notif: Handle value indication packet
        :type notif: ATT_Handle_Value_Indication
        """
        self.send('gatt', GattHandleValueIndication(
                notif.gatt_handle,
                notif.value
            )
        )

    ##########################################
    # Outgoing requests and responses
    ##########################################

    def send_data(self, packet: Packet):
        """Send packet to underlying L2CAP layer

        :param packet: Packet to send to L2CAP
        :type packet: Packet
        """
        self.send('l2cap', ATT_Hdr()/packet)

    def error_response(self, request, handle, reason):
        """Sends an ATT Error Response

        :param int request: Request that generated this error
        :param int handle: Attribute handle that generated this error
        :param int ecode: Reason why this error has been generated
        """
        self.send_data(ATT_Error_Response(
                request=request,
                handle=handle,
                ecode=reason
            )
        )

    def exch_mtu_request(self, mtu):
        """Sends an ATT Exchange MTU Request

        :param int mtu: Maximum Transmission Unit
        """
        # Update local MTU first
        logger.debug("[att] sending an MTU exchange request (mtu: %d)", mtu)
        self.send_data(ATT_Exchange_MTU_Request(
            mtu=mtu
        ))

    def exch_mtu_response(self, mtu):
        """Sends an ATT Exchange MTU Response

        :param int mtu: Maximum Transmission Unit
        """
        logger.debug("[att] sending an MTU exchange response (mtu: %d)", mtu)
        self.send_data(ATT_Exchange_MTU_Response(
            mtu=mtu
        ))


    def find_info_request(self, start, end):
        """Sends an ATT Find Information Request
        """
        self.send_data(ATT_Find_Information_Request(
            start=start,
            end=end
        ))

    def  find_info_response(self, form, handles):
        """Sends an ATT Find Information Response
        """

        self.send_data(ATT_Find_Information_Response(
            format=form,
            handles=handles
        ))

    def find_by_type_value_request(self, start, end, type_uuid, value):
        """Sends an ATT Find By Type Value Request
        """
        self.send_data(ATT_Find_By_Type_Value_Request(
            start=start,
            end=end,
            uuid=type_uuid,
            data=value
        ))

    def  find_by_type_value_response(self, handles: List[ATT_Handle]):
        """Sends an ATT Find By Type Value Response
        """
        self.send_data(ATT_Find_By_Type_Value_Response(
            handles=handles
        ))

    def read_by_type_request(self, start, end, uuid):
        """Sends an ATT Read By Type Request

        :param int start: First requested handle number
        :param int end: Last requested handle number
        :param uuid: 16-bit or 128-bit attribute UUID
        """
        self.send_data(ATT_Read_By_Type_Request(
            start=start,
            end=end,
            uuid=uuid
        ))


    def read_by_type_request_128bit(self, start, end, uuid1, uuid2):
        """Sends an ATT Read By Type Request with 128-bit UUID

        :param int start: First requested handle number
        :param int end: Last requested handle number
        :param uuid1: UUID part 1
        :param uuid2: UUID part 2
        """
        self.send_data(ATT_Read_By_Type_Request_128bit(
            start=start,
            end=end,
            uuid1=uuid1,
            uuid2=uuid2
        ))

    def read_by_type_response(self, item_length, handles):
        """Sends an ATT Read By Type Response

        :param int item_length: Length of a handle item
        :param list handles: List of handles (each item stored on `item_length` bytes)
        """
        self.send_data(ATT_Read_By_Type_Response(
            len=item_length,
            handles=handles
        ))


    def read_request(self, gatt_handle):
        """Sends an ATT Read Request
        """
        self.send_data(ATT_Read_Request(
            gatt_handle=gatt_handle
        ))

    def read_response(self, value):
        """Sends an ATT Read Response
        """
        self.send_data(ATT_Read_Response(
            value=value
        ))


    def read_blob_request(self, handle, offset):
        """Sends an ATT Read Blob Request

        :param int handle: Handle of attribute to read from
        :param int offset: Offset of the first octet to be read
        """
        self.send_data(ATT_Read_Blob_Request(
            gatt_handle=handle,
            offset=offset
        ))

    def read_blob_response(self, value):
        """Sends an ATT Read Blob Response

        :param value: Value read
        """
        self.send_data(ATT_Read_Blob_Response(
            value=value
        ))

    def read_multiple_request(self, handles):
        """Sends an ATT Read Multiple Request

        :param handles: list of handles
        """
        self.send_data(ATT_Read_Multiple_Request(
            handles=handles
        ))

    def read_multiple_response(self, values):
        """Sends an ATT Read Multiple Response

        :param values: List of multiple values
        """
        self.send_data(ATT_Read_Multiple_Response(
            values=values
        ))


    def read_by_group_type_request(self, start, end, uuid):
        """Sends an ATT Read By Group Type Request

        :param int start: First requested handle number
        :param int end: Last requested handle number
        :param uuid: 16-bit or 128-bit group UUID
        """
        self.send_data(ATT_Read_By_Group_Type_Request(
            start=start,
            end=end,
            uuid=uuid
        ))

    def read_by_group_type_response(self, length, data):
        """Sends an ATT Read By Group Type Response

        :param int length: Size of each attribute data
        :param data: List of attribute data
        """
        self.send_data(ATT_Read_By_Group_Type_Response(
            length=length,
            data=data
        ))

    def write_request(self, handle, data):
        """Sends an ATT Write Request

        :param int handle: Attribute handle to write into
        :param data: Data to write
        """
        self.send_data(ATT_Write_Request(
            gatt_handle=handle,
            data=data
        ))

    def write_response(self):
        """Sends an ATT Write Response
        """
        self.send_data(ATT_Write_Response())

    def write_command(self, handle, data):
        """Sends an ATT Write Command
        """
        self.send_data(ATT_Write_Command(
            gatt_handle=handle,
            data=data
        ))


    def prepare_write_request(self, handle, offset, data):
        """Sends an ATT Write Request

        :param int handle: Attribute handle
        :param int offset: Offset of the data to write
        :param data: Data to write
        """
        self.send_data(ATT_Prepare_Write_Request(
            gatt_handle=handle,
            offset=offset,
            data=data
        ))

    def prepare_write_response(self, handle, offset, data):
        """Sends an ATT Write Response

        :param int handle: Attribute handle
        :param int offset: Offset of the data to write
        :param data: Data to write
        """
        self.send_data(ATT_Prepare_Write_Response(
            gatt_handle=handle,
            offset=offset,
            data=data
        ))

    def execute_write_request(self, flags):
        """Sends an ATT Execute Write Request

        :param flags: Flags
        """
        self.send_data(ATT_Execute_Write_Request(
            flags=flags
        ))

    def execute_write_response(self):
        """Sends an ATT Execute Write Response
        """
        self.send_data(ATT_Execute_Write_Response())


    def handle_value_notification(self, handle, value):
        """Sends an ATT Handle Value Notification

        :param int handle: Attribute handle
        :param value: Attribute value
        """
        self.send_data(ATT_Handle_Value_Notification(
            gatt_handle=handle,
            value=value
        ))

    def handle_value_indication(self, handle, value):
        """Sends an ATT Handle Value Indication

        :param int handle: Attribute handle
        :param value: Attribute value
        """
        self.send_data(ATT_Handle_Value_Indication(
            gatt_handle=handle,
            value=value
        ))

    def handle_value_confirmation(self):
        """Sends an ATT Handle Value Confirmation

        Not supported yet
        """
        self.send_data(ATT_Handle_Value_Confirmation())



"""WHAD Protocol Generic Debug message abstraction layer.
"""
from whad.hub.message import pb_bind, PbFieldBytes, PbMessageWrapper
from whad.hub.generic import Generic

@pb_bind(Generic, 'verbose', 1)
class Verbose(PbMessageWrapper):
    """Generic verbose message.
    """
    msg = PbFieldBytes('generic.verbose.data')



from scapy.packet import Packet, bind_layers, Raw
from scapy.config import conf
from scapy.fields import ByteField, ByteEnumField, LEShortEnumField, \
    FieldLenField, PacketListField, LenField, LEThreeBytesField, \
    IEEEFloatField, LEIntField, LEShortField, IEEEDoubleField, StrFixedLenField, Field
from scapy.layers.dot15d4 import Dot15d4, Dot15d4FCS
import struct

class LEIEEEFloatField(Field[int, int]):
    def __init__(self, name, default):
        Field.__init__(self, name, default, "<f")


class Dot15d4Raw(Raw):
    """IEEE 802.15.4 raw packet
    """

# Implement https://github.com/jkcko/ieee802.15.4-tap
DLT_IEEE802_15_4_TAP = 283

DOT15D4TAP_VERSIONS = {
    0x00 : "version 1.0"
}

DOT15D4TAP_TYPES = {
    0x00 : "fcs_type",
    0x01 : "receive_signal_strength",
    0x02 : "bit_rate",
    0x03 : "channel_assignment",
    0x04 : "sun_phy_information",
    0x05 : "start_of_frame_timestamp",
    0x06 : "end_of_frame_timestamp",
    0x07 : "absolute_slot_number",
    0x08 : "start_of_slot_timestamp",
    0x09 : "timeslot_length",
    0x0a : "link_quality_indicator",
    0x0b : "channel_center_frequency",
    0x0c : "channel_plan"

}

class Dot15d4TAP_TLV_Hdr(Packet):
    name = "Dot15d4 TAP Type-Length-Value Header"
    fields_desc = [
        LEShortEnumField("type", None, DOT15D4TAP_TYPES),
        LenField("len", None, fmt="<H")
    ]


class Dot15d4TAP_TLV_Element(Packet):
    name = "Dot15d4 TAP Type-Length-Value Element"
    def extract_padding(self, s):
        return b"",s

class  Dot15d4TAP_FCS_Type(Dot15d4TAP_TLV_Element):
    name = "Dot15d4 FCS Type"
    fields_desc = [
        ByteEnumField("fcs_type", 1, {0 : "none", 1: "16-bit CRC", 2: "32-bit CRC"}),
        StrFixedLenField("padding", b"\x00"*3, 3)

    ]

class  Dot15d4TAP_Received_Signal_Strength(Dot15d4TAP_TLV_Element):
    name = "Dot15d4 Received Signal Strength"
    fields_desc = [
        LEIEEEFloatField("rss",None),
    ]

class  Dot15d4TAP_Bit_Rate(Dot15d4TAP_TLV_Element):
    name = "Dot15d4 Bit Rate"
    fields_desc = [
        LEIntField("bit_rate",None)
    ]

class  Dot15d4TAP_Channel_Assignment(Dot15d4TAP_TLV_Element):
    name = "Dot15d4 Channel Assignment"
    fields_desc = [
        LEShortField("channel_number",None),
        ByteField("channel_page", None),
        StrFixedLenField("padding", "\x00", 1),
    ]

class  Dot15d4TAP_SUN_PHY_Information(Dot15d4TAP_TLV_Element):
    name = "Dot15d4 SUN PHY Information"
    fields_desc = [
        ByteField("phy_band", None),
        ByteField("phy_type", None),
        ByteField("phy_mode", None),
        StrFixedLenField("padding", b"\x00"*1, 1)
    ]


class  Dot15d4TAP_Start_Of_Frame_Timestamp(Dot15d4TAP_TLV_Element):
    name = "Dot15d4 Start of Frame Timestamp"
    fields_desc = [
        IEEEDoubleField("timestamp", None)
    ]


class  Dot15d4TAP_End_Of_Frame_Timestamp(Dot15d4TAP_TLV_Element):
    name = "Dot15d4 End of Frame Timestamp"
    fields_desc = [
        IEEEDoubleField("timestamp", None)
    ]

class  Dot15d4TAP_Absolute_Slot_Number(Dot15d4TAP_TLV_Element):
    name = "Dot15d4 Absolute Slot Number"
    fields_desc = [
        IEEEDoubleField("absolute_slot_number", None)
    ]


class  Dot15d4TAP_Start_Of_Slot_Timestamp(Dot15d4TAP_TLV_Element):
    name = "Dot15d4 Start of Slot Timestamp"
    fields_desc = [
        IEEEDoubleField("timestamp", None)
    ]

class  Dot15d4TAP_Timeslot_Length(Dot15d4TAP_TLV_Element):
    name = "Dot15d4 Timeslot Length"
    fields_desc = [
        IEEEDoubleField("timeslot_length", None)
    ]


class  Dot15d4TAP_Link_Quality_Indicator(Dot15d4TAP_TLV_Element):
    name = "Dot15d4 Link Quality Indicator"
    fields_desc = [
        ByteField("lqi", None),
        StrFixedLenField("padding", b"\x00"*3, 3)
    ]


class  Dot15d4TAP_Channel_Center_Frequency(Dot15d4TAP_TLV_Element):
    name = "Dot15d4 Channel Center Frequency"
    fields_desc = [
        LEIEEEFloatField("channel_frequency", None),
    ]


class  Dot15d4TAP_Channel_Plan(Dot15d4TAP_TLV_Element):
    name = "Dot15d4 Channel Plan"
    fields_desc = [
        LEIEEEFloatField("base_channel_frequency", None),
        LEIEEEFloatField("channel_spacing", None),
        LEShortField("number_of_channels", None),
        StrFixedLenField("padding", b"\x00"*2, 2)
    ]


class Dot15d4TAP_Hdr(Packet):
    name = "Dot15d4 TAP Header"
    fields_desc = [
        ByteEnumField("version", 0, DOT15D4TAP_VERSIONS),
        ByteField("reserved", 0),
        FieldLenField("length", None, length_of="data", fmt="<H", adjust=lambda pkt, f:f+4),
        PacketListField("data", [], Dot15d4TAP_TLV_Hdr, length_from=lambda x:x.length-4),
    ]


bind_layers(Dot15d4TAP_TLV_Hdr, Dot15d4TAP_FCS_Type, type=0x00)
bind_layers(Dot15d4TAP_TLV_Hdr, Dot15d4TAP_Received_Signal_Strength, type=0x01)
bind_layers(Dot15d4TAP_TLV_Hdr, Dot15d4TAP_Bit_Rate, type=0x02)
bind_layers(Dot15d4TAP_TLV_Hdr, Dot15d4TAP_Channel_Assignment, type=0x03)
bind_layers(Dot15d4TAP_TLV_Hdr, Dot15d4TAP_SUN_PHY_Information, type=0x04)
bind_layers(Dot15d4TAP_TLV_Hdr, Dot15d4TAP_Start_Of_Frame_Timestamp, type=0x05)
bind_layers(Dot15d4TAP_TLV_Hdr, Dot15d4TAP_End_Of_Frame_Timestamp, type=0x06)
bind_layers(Dot15d4TAP_TLV_Hdr, Dot15d4TAP_Absolute_Slot_Number, type=0x07)
bind_layers(Dot15d4TAP_TLV_Hdr, Dot15d4TAP_Start_Of_Slot_Timestamp, type=0x08)
bind_layers(Dot15d4TAP_TLV_Hdr, Dot15d4TAP_Timeslot_Length, type=0x09)
bind_layers(Dot15d4TAP_TLV_Hdr, Dot15d4TAP_Link_Quality_Indicator, type=0x0a)
bind_layers(Dot15d4TAP_TLV_Hdr, Dot15d4TAP_Channel_Center_Frequency, type=0x0b)
bind_layers(Dot15d4TAP_TLV_Hdr, Dot15d4TAP_Channel_Plan, type=0x0c)

bind_layers(Dot15d4TAP_Hdr, Dot15d4)
bind_layers(Dot15d4TAP_Hdr, Dot15d4FCS)
conf.l2types.register(DLT_IEEE802_15_4_TAP, Dot15d4TAP_Hdr)



"""
Pythonic 802.15.4 stack
"""

from whad.protocol.dot15d4.dot15d4_pb2 import AddressType
from whad.dot15d4.utils.phy import PHYS
from whad.dot15d4.stack.mac import MACManager

from whad.common.stack import Layer, alias, source

import logging
logger = logging.getLogger(__name__)

@alias('phy')
class Dot15d4Stack(Layer):
    """
    This class holds the main components of a (hackable) 802.15.4 stack:

    - the Medium Access Control Manager (MAC - defined in 802.15.4 specification)

    The Medium Access Control manager handles all the low-level operations:
    - 802.15.4 management control (handles beaconing, frame validation, timeslots and associations)
    - 802.15.4 data (forward to upper layer, i.e. NWK)
    """

    def __init__(self, connector, phy=PHYS["802.15.4-OQPSK"], options={}):
        super().__init__(options=options)

        # Save connector (used as PHY layer)
        self.__connector = connector

        # Save selected physical layer
        self.__selected_phy = phy


    @property
    def symbol_duration(self):
        """
        Compute symbol duration (in us).
        """
        return (4 / self.__selected_phy.datarate)

    #############################
    # Incoming messages
    #############################
    def on_pdu(self, pdu):
        '''PDU callback.

        This callback handles a PDU and forwards it to MAC layer.
        '''
        logger.debug("received a PDU (%d bytes)", len(pdu))
        self.send('mac', pdu, tag='pdu')

    def on_ed_sample(self, timestamp, sample):
        '''Energy Detection sample callback.

        This callback handle an Energy Detection sample and forwards it to MAC layer.
        '''
        logger.debug("received an energy detection sample (%d / timestamp=%d) ", sample, timestamp)

        # Notify link layer we received a control PDU for a given `conn_handle`.
        self.send('mac', sample, tag='energy_detection', timestamp=timestamp)

    ############################
    # Interact
    ############################

    def set_short_address(self, address):
        '''Select short address.
        '''
        self.__connector.set_node_address(address, mode=AddressType.SHORT)

    def set_extended_address(self, address):
        '''Select extended address.
        '''
        self.__connector.set_node_address(address, mode=AddressType.EXTENDED)

    def set_channel(self, channel):
        '''Select channel.
        '''
        self.__connector.set_channel(channel)

    def set_channel_page(self, page):
        '''Select channel page.
        '''
        self.__connector.set_channel_page(page)

    def get_channel(self):
        '''Get channel.
        '''
        return self.__connector.get_channel()

    def get_channel_page(self):
        '''Get channel page.
        '''
        return self.__connector.get_channel_page()

    def perform_ed_scan(self, channel):
        '''Start an Energy Detection scan.
        '''
        self.__connector.perform_ed_scan(channel)

    @source('mac', 'pdu')
    def transmit(self, packet):
        '''Send PDU to the underlying WHAD connector.
        '''
        self.__connector.send(packet)

Dot15d4Stack.add(MACManager)



"""WHAD server tool

This utility implements a server module, allowing to create a TCP proxy
which can be used to access a device remotely.
"""
import logging

from prompt_toolkit import print_formatted_text, HTML

from whad.cli.app import CommandLineApp, run_app
from whad.device.tcp import TCPSocketConnector
from whad.device.websocket import WebSocketConnector

logger = logging.getLogger(__name__)

class WhadServerApp(CommandLineApp):
    """Main wserver CLI application class.
    """

    def __init__(self):
        """Application uses an interface and has commands.
        """
        super().__init__(
            description="WHAD server tool",
            interface=True,
            commands=False
        )

        self.add_argument(
            '--websocket',
            '--ws',
            '-w',
            action='store_true',
            dest='websocket',
            default=False,
            help="Enable websocket mode."
        )


        self.add_argument(
            '-j',
            '--json',
            action='store_true',
            dest='json',
            default=False,
            help="Export as JSON (websocket mode only)"
        )

        self.add_argument(
            "--address",
            "-a",
            dest="address",
            action="store",
            default="127.0.0.1",
            help="IP address to use"
        )

        self.add_argument(
            "--port",
            "-p",
            dest="port",
            action="store",
            default="12345",
            help="Port to use"
        )

        # Initialize properties.
        self.address = None
        self.port = None
        self.server = None

    def run(self, pre: bool = True, post: bool = True):
        """CLI application main routine.
        """
        try:
            # Launch pre-run tasks
            self.pre_run()

            if self.is_piped_interface():
                interface = self.input_interface
            else:
                interface = self.interface

            # Format address and port
            self.address = self.args.address
            self.port = int(self.args.port)

            # We need to have an interface specified
            if interface is not None:
                self.serve(interface)
            else:
                self.error("You have to provide an interface to proxify.")

        except KeyboardInterrupt:
            self.warning("Server stopped (CTRL-C)")

        if self.server is not None:
            self.server.shutdown()

        # Launch post-run tasks
        self.post_run()

    def serve(self, device):
        """
        Create a TCP proxy device according to provided address and port and serve forever.
        """
        print_formatted_text(HTML(
            f"<ansicyan>[i] Device proxy running on {self.address}:{self.port} </ansicyan>"
        ))

        if self.args.websocket:
            logger.debug("[wserver] Uses websocket mode")
            self.server = WebSocketConnector(device, self.address, self.port,
                                             json_mode = self.args.json)
            self.server.serve()
        else:
            logger.debug("[wserver] Uses default TCP mode")
            # Setup a TCP server and await connections.
            self.server = TCPSocketConnector(device, self.address, self.port)
            self.server.serve()

def wserver_main():
    """Launcher for wserver.
    """
    app = WhadServerApp()
    run_app(app)



"""WHAD Protocol Bluetooth Low Energy domain message abstraction layer.
"""
from typing import List, Optional
from dataclasses import dataclass, field, fields

from .bdaddr import BDAddress
from .chanmap import ChannelMap

from scapy.layers.bluetooth4LE import BTLE_RF, BTLE, BTLE_ADV, BTLE_DATA

from whad.protocol.ble.ble_pb2 import BleDirection, BleAdvType, BleAddrType
from whad.hub.registry import Registry
from whad.hub.message import HubMessage, pb_bind
from whad.hub import ProtocolHub
from whad.hub.metadata import Metadata

class Commands:
    """BLE Commands
    """
    SetBdAddress = 0x00
    SniffAdv = 0x01
    JamAdv = 0x02
    JamAdvOnChannel = 0x03
    ReactiveJam = 0x04
    SniffConnReq = 0x05
    SniffAccessAddress = 0x06
    SniffActiveConn = 0x07
    JamConn = 0x08
    ScanMode = 0x09
    AdvMode = 0x0a
    SetAdvData = 0x0b
    CentralMode = 0x0c
    ConnectTo = 0x0d
    SendRawPDU = 0x0e
    SendPDU = 0x0f
    Disconnect = 0x10
    PeripheralMode = 0x11
    Start = 0x12
    Stop = 0x13
    SetEncryption = 0x14
    HijackMaster = 0x15
    HijackSlave = 0x16
    HijackBoth = 0x17
    PrepareSequence = 0x18
    TriggerSequence = 0x19
    DeleteSequence = 0x1a


class Direction:
    """BLE PDU direction
    """
    UNKNOWN = BleDirection.UNKNOWN
    MASTER_TO_SLAVE = BleDirection.MASTER_TO_SLAVE
    SLAVE_TO_MASTER = BleDirection.SLAVE_TO_MASTER
    INJECTION_TO_SLAVE = BleDirection.INJECTION_TO_SLAVE
    INJECTION_TO_MASTER = BleDirection.INJECTION_TO_MASTER

class AdvType:
    ADV_UNKNOWN = BleAdvType.ADV_UNKNOWN
    ADV_IND = BleAdvType.ADV_IND
    ADV_DIRECT_IND = BleAdvType.ADV_DIRECT_IND
    ADV_NONCONN_IND = BleAdvType.ADV_NONCONN_IND
    ADV_SCAN_IND = BleAdvType.ADV_SCAN_IND
    ADV_SCAN_RSP = BleAdvType.ADV_SCAN_RSP

class AddressType:
    PUBLIC = BleAddrType.PUBLIC
    RANDOM = BleAddrType.RANDOM
    
@dataclass(repr=False)
class BLEMetadata(Metadata):
    direction : BleDirection = None
    connection_handle : int = None
    is_crc_valid : bool = None
    relative_timestamp : int = None
    decrypted : bool = None
    encrypt: bool = False

    @classmethod
    def convert_from_header(cls, pkt):
        header = pkt[BTLE_RF]
        if header.type == 2:
            direction = BleDirection.MASTER_TO_SLAVE
        elif header.type == 3:
            direction = BleDirection.SLAVE_TO_MASTER
        else:
            direction = BleDirection.UNKNOWN

        channel = header.rf_channel
        is_crc_valid = header.crc_valid == 1
        rssi = header.signal

        return BLEMetadata(
            direction = direction,
            is_crc_valid = is_crc_valid,
            rssi = rssi,
            channel = channel,
            timestamp = int(100000 * pkt.time)
        )

    def convert_to_header(self):
        timestamp = None
        packet_type = 0 # ADV_OR_DATA_UNKNOWN_DIR
        signal = -128
        crc_checked = 0
        crc_valid = 0
        sig_power_valid = 0
        dewhitened = 1
        rf_channel = 0
        if self.direction is not None:
            if self.direction == BleDirection.MASTER_TO_SLAVE:
                packet_type = 2
            elif self.direction == BleDirection.SLAVE_TO_MASTER:
                packet_type = 3
        if self.timestamp is not None:
            timestamp = self.timestamp
        if self.rssi is not None:
            sig_power_valid = 1
            signal = self.rssi
        if self.is_crc_valid is not None:
            crc_checked = 1
            crc_valid = self.is_crc_valid
        if self.channel is not None:
            rf_channel = self.channel

        header = BTLE_RF(
            rf_channel = rf_channel,
            type = packet_type,
            signal = signal,
            crc_checked = crc_checked,
            crc_valid = crc_valid,
            sig_power_valid = sig_power_valid,
            dewhitened = dewhitened
        )
        return header, timestamp

def generate_ble_metadata(message):
    metadata = BLEMetadata()
    if isinstance(message, BleRawPduReceived):
        metadata.direction = message.direction
        if message.rssi is not None:
            metadata.rssi = message.rssi
        metadata.channel = message.channel
        if message.timestamp is not None:
            metadata.timestamp = message.timestamp
        if message.crc_validity is not None:
            metadata.is_crc_valid = message.crc_validity
        if message.relative_timestamp is not None:
            metadata.relative_timestamp = message.relative_timestamp
            metadata.decrypted = message.decrypted

        metadata.connection_handle = message.conn_handle

    elif isinstance(message, BleAdvPduReceived):
        metadata.direction = BleDirection.UNKNOWN
        metadata.rssi = message.rssi

    elif isinstance(message, BlePduReceived):
        metadata.connection_handle = message.conn_handle
        metadata.direction = message.direction
        metadata.decrypted = message.decrypted

    elif isinstance(message, SendBlePdu):
        metadata.connection_handle = message.conn_handle
        metadata.direction = message.direction

    elif isinstance(message, SendBleRawPdu):
        metadata.direction = message.direction
        metadata.crc = message.crc
        metadata.connection_handle = message.conn_handle

    return metadata

@pb_bind(ProtocolHub, name="ble", version=1)
class BleDomain(Registry):
    """WHAD BLE domain messages parser/factory.
    """

    NAME = 'ble'
    VERSIONS = {}

    def __init__(self, version: int):
        """Initializes a BLE domain instance
        """
        self.proto_version = version

    @staticmethod
    def parse(proto_version: int, message) -> HubMessage:
        """Parses a WHAD BleDomain message as seen by protobuf
        """
        message_type = message.ble.WhichOneof('msg')
        message_clazz = BleDomain.bound(message_type, proto_version)
        return message_clazz.parse(proto_version, message)

    def is_packet_compat(self, packet) -> bool:
        """Determine if a packet is a compatible BLE packet
        """
        return isinstance(packet.metadata, BLEMetadata)

    def convert_packet(self, packet) -> Optional[HubMessage]:
        """Convert a BLE packet to SendPdu or SendBlePdu message.
        """
        if isinstance(packet.metadata, BLEMetadata):
            if packet.metadata.raw:
                return BleDomain.bound('send_raw_pdu', self.proto_version).from_packet(
                    packet, encrypt=packet.metadata.encrypt
                )
            else:
                return BleDomain.bound('send_pdu', self.proto_version).from_packet(
                    packet, encrypt=packet.metadata.encrypt
                )
        else:
            # Error
            return None

    def format(self, packet):
        """Convert this message to its scapy representation with the
        appropriate header and timestamp in microseconds.
        """
        formatted_packet = packet
        if BTLE not in packet:
            if BTLE_ADV in packet:
                formatted_packet = BTLE(access_addr=0x8e89bed6)/packet
            elif BTLE_DATA in packet:
                # We are forced to use a pseudo access address for connections in this case.
                formatted_packet = BTLE(access_addr=0x11223344) / packet

        timestamp = None
        if hasattr(packet, "metadata"):
            header, timestamp = packet.metadata.convert_to_header()
            formatted_packet = header / formatted_packet
        else:
            header = BTLE_RF()
            formatted_packet = header / formatted_packet

        return formatted_packet, timestamp

    def create_set_bd_address(self, bd_address: BDAddress) -> HubMessage:
        """Create a SetBdAddress message.

        :param bd_address: BD address to set
        :type bd_address: BDAddress
        :return: SetBdAddress message
        :rtype: HubMessage
        """
        return BleDomain.bound('set_bd_addr', self.proto_version)(
            bd_address=bd_address.value,
            addr_type=AddressType.PUBLIC if bd_address.is_public() else AddressType.RANDOM
        )

    def create_sniff_adv(self, channel: int, bd_address: BDAddress = None,
                       use_ext_adv: bool = False) -> HubMessage:
        """Create a SniffAdv message.

        :param channel: BLE channel to listen on
        :type channel: int
        :param bd_address: BD address to target, if given
        :type bd_address: BDAddress, optional
        :param use_ext_adv: Use extended advertisements
        :type use_ext_adv: bool, optional
        :return: instance of SniffAdv message
        :rtype: HubMessage
        """
        if bd_address is not None:
            target_address = bd_address
        else:
            target_address = BDAddress('FF:FF:FF:FF:FF:FF')
        return BleDomain.bound('sniff_adv', self.proto_version)(
            bd_address=target_address.value,
            channel=channel,
            use_extended_adv=use_ext_adv
        )

    def create_sniff_connreq(self, channel: int, bd_address: BDAddress = None,
                           show_empty: bool = False, show_adv: bool = False) -> HubMessage:
        """Create a SniffConnReq message.

        :param channel: BLE channel to listen on
        :type channel: int
        :param bd_address: BD address to target, if given
        :type bd_address: BDAddress, optional
        :param show_empty: Report empty PDUs
        :type show_empty: bool, optional
        :param show_adv: Report advertisements
        :type show_adv: bool, optional
        :rtype: HubMessage
        """
        if bd_address is not None:
            target_address = bd_address
        else:
            target_address = BDAddress('FF:FF:FF:FF:FF:FF')
        return BleDomain.bound('sniff_connreq', self.proto_version)(
            bd_address=target_address.value,
            channel=channel,
            show_empty_packets=show_empty,
            show_advertisements=show_adv
        )

    def create_sniff_access_address(self, channels: List[int]) -> HubMessage:
        """Create a SniffAccessAddress message.

        :param channels: List of channels
        :type channels: list
        :return: an instance of SniffAccessAddress message
        :rtype: HubMessage
        """
        return BleDomain.bound('sniff_aa', self.proto_version)(
            monitored_channels=ChannelMap(channels).value
        )

    def create_sniff_active_conn(self, access_address: int, crc_init: int = None,
                              channel_map: ChannelMap = None, interval: int = None,
                              increment: int = None, channels: List[int] = None):
        """Create a SniffActiveConn message.

        :param access_address: Target connection access address
        :type access_address: int
        :param crc_init: Connection CRC initial seed value
        :type crc_init: int, optional
        :param channel_map: Channel map to use when sniffing connection
        :type channel_map: ChannelMap, optional
        :param interval: Hop interval to use when sniffing connection
        :type interval: int, optional
        :param increment: Hop increment to use when sniffing connection
        :type increment: int, optional
        :param channels: Channels to sniff on when recovering connection parameters
        :type channels: list, optional
        :return: instance of SniffActiveConn message
        :rtype: SniffActiveConn
        """
        # Create default createSniffActiveConn message
        sniff_connreq = BleDomain.bound('sniff_conn', self.proto_version)(
            access_address=access_address
        )

        # Add optional fields if provided
        if crc_init is not None:
            sniff_connreq.crc_init = crc_init
        if channel_map is not None:
            sniff_connreq.channel_map = channel_map.value
        if interval is not None:
            sniff_connreq.hop_interval = interval
        if increment is not None:
            sniff_connreq.hop_increment = increment
        if channels is not None:
            sniff_connreq.monitored_channels = bytes(channels)

        # Return the created SniffConnReq message
        return sniff_connreq

    def create_access_address_discovered(self, access_address: int, rssi: int, timestamp: int) -> HubMessage:
        """Create an AccessAddressDiscovered notification message.

        :param access_address: Discovered access address
        :type access_address: int
        :param rssi: Received Signal Strength Indicator
        :type rssi: int
        :param timestamp: Timestamp at which the access address has been discovered
        :type timestamp: int
        :return: instance of AccessAddressDiscovered
        :rtype: AccessAddressDiscovered
        """
        return BleDomain.bound('aa_disc', self.proto_version)(
            access_address=access_address,
            rssi=rssi,
            timestamp=timestamp
        )

    def create_jam_adv(self) -> HubMessage:
        """Create a JamAdv message.

        :return: instance of JamAdv message
        :rtype: JamAdv
        """
        return BleDomain.bound('jam_adv', self.proto_version)()

    def create_jam_adv_chan(self, channel: int) -> HubMessage:
        """Create a JamAdvChan message.

        :param channel: Advertising channel to jam
        :type channel: int
        :return: instance of JamAdvChan
        :rtype: JamAdvChan
        """
        return BleDomain.bound('jam_adv_chan', self.proto_version)(
            channel=channel
        )

    def create_jam_conn(self, access_address: int) -> HubMessage:
        """Create a JamConn message.

        :param access_address: Target connection access address
        :type access_address: int
        :return: instance of JamConn message
        :rtype: JamConn
        """
        return BleDomain.bound('jam_conn', self.proto_version)(
            access_address=access_address
        )

    def create_reactive_jam(self, channel: int, pattern: bytes, position: int) -> HubMessage:
        """Create a ReactiveJam message.

        :param channel: Target channel
        :type channel: int
        :param pattern: Trigger pattern
        :type pattern: bytes
        :param position: Triger pattern position
        :type position: int
        :return: instance of ReactiveJam
        :rtype: ReactiveJam
        """
        return BleDomain.bound('reactive_jam', self.proto_version)(
            channel=channel,
            pattern=pattern,
            position=position
        )

    def create_scan_mode(self, active: bool = False) -> HubMessage:
        """Create a ScanMode message.

        :param active: Enable active scan mode
        :type active: bool
        :return: instance of ScanMode
        :rtype: ScanMode
        """
        return BleDomain.bound('scan_mode', self.proto_version)(
            active=active
        )

    def create_adv_mode(self, adv_data: bytes, scan_rsp: bytes = None) -> HubMessage:
        """Create an AdvMode message.

        :param adv_data: Advertisement data (31 bytes max)
        :type adv_data: bytes
        :param scan_rsp: Scan response data (31 bytes max)
        :type scan_rsp: bytes, optional
        :return: instance of AdvMode message
        :rtype: AdvMode
        """
        message = BleDomain.bound('adv_mode', self.proto_version)(
            scan_data=adv_data
        )
        if scan_rsp is not None:
            message.scanrsp_data = scan_rsp
        return message

    def create_central_mode(self) -> HubMessage:
        """Create a CentralMode message.

        :return: instance of CentralMode message
        :rtype: CentralMode
        """
        return BleDomain.bound('central_mode', self.proto_version)()

    def create_periph_mode(self, adv_data: bytes = None, scan_rsp: bytes = None) -> HubMessage:
        """Create an PeriphMode message.

        :param adv_data: Advertisement data (31 bytes max)
        :type adv_data: bytes
        :param scan_rsp: Scan response data (31 bytes max)
        :type scan_rsp: bytes, optional
        :return: instance of PeriphMode message
        :rtype: PeriphMode
        """
        message = BleDomain.bound('periph_mode', self.proto_version)(
        )
        if adv_data is not None:
            message.scan_data = adv_data
        if scan_rsp is not None:
            message.scanrsp_data = scan_rsp
        return message

    def create_start(self) -> HubMessage:
        """Create a Start message.

        :return: instance of Start message
        :rtype: Start
        """
        return BleDomain.bound("start", self.proto_version)()

    def create_stop(self) -> HubMessage:
        """Create a Stop message.

        :return: instance of Stop message
        :rtype: Stop
        """
        return BleDomain.bound("stop", self.proto_version)()

    def create_connect_to(self, bd_address: BDAddress = None, access_address: int = None,
                        channel_map: ChannelMap = None, interval: int = None,
                        increment: int = None, crc_init: int = None) ->HubMessage:
        """Create a ConnectTo message.

        :param bd_address: Target BD address
        :type bd_address: BDAddress, optional
        :param access_address: Access address to synchronize with
        :type access_address: int, optional
        :param channel_map: Channel map to use with the connection
        :type channel_map: ChannelMap, optional
        :param interval: Hop interval to use
        :type interval: int, optional
        :param increment: Hop increment to use
        :type increment: int, optional
        :param crc_init: CRC initial value to use
        :type crc_init: int, optional
        :return: instance of ConnectTo message
        :rtype: ConnectTo
        """
        message = BleDomain.bound("connect", self.proto_version)()

        # Set bd address if provided
        if bd_address is not None:
            message.bd_address = bd_address.value
            message.addr_type = AddressType.PUBLIC if bd_address.is_public() else AddressType.RANDOM

        # Set access address if provided
        if access_address is not None:
            message.access_address = access_address

        # Set channel map if provided
        if channel_map is not None:
            message.channel_map = channel_map.value

        # Set hop interval if provided
        if interval is not None:
            message.hop_interval = interval

        # Set hop increment if provided
        if increment is not None:
            message.hop_increment = increment

        # Set crc init
        if crc_init is not None:
            message.crc_init = crc_init

        return message


    def create_disconnect(self, conn_handle: int) -> HubMessage:
        """Create a Disconnect message.

        :param conn_handle: Connection handle
        :type conn_handle: int
        :return: instance of Disconnect message
        :rtype: Disconnect
        """
        return BleDomain.bound("disconnect", self.proto_version)(
            conn_handle=conn_handle
        )

    def create_synchronized(self, access_address: int, interval: int, increment: int,\
                           channel_map: ChannelMap, crc_init: int) -> HubMessage:
        """Create a Synchronized message.

        :param access_address: Connection access address
        :type access_address: int
        :param interval: Connection hop interval
        :type interval: int
        :param increment: Connection hop increment
        :type increment: int
        :param channel_map: Connection channel map
        :type channel_map: ChannelMap
        :return: instance of Synchronized
        :rtype: Synchronized
        """
        return BleDomain.bound("synchronized", self.proto_version)(
            access_address=access_address,
            crc_init=crc_init,
            hop_interval=interval,
            hop_increment=increment,
            channel_map=channel_map.value
        )

    def create_connected(self, initiator: BDAddress, advertiser: BDAddress, \
                        access_address: int, conn_handle: int) -> HubMessage:
        """Create a Connected message.

        :param initiator: Connection initiator BD address
        :type initiator: BDAddress
        :param advertiser: Connection advertiser BD address
        :type advertiser: BDAddress
        :param access_address: Connection access address
        :type access_address: int
        :param conn_handle: Connection handle
        :type conn_handle: int
        :return: instance of a Connected message
        :rtype: Connected
        """
        return BleDomain.bound("connected", self.proto_version)(
            initiator=initiator.value,
            advertiser=advertiser.value,
            access_address=access_address,
            conn_handle=conn_handle,
            adv_addr_type=AddressType.PUBLIC if advertiser.is_public() else AddressType.RANDOM,
            init_addr_type=AddressType.PUBLIC if initiator.is_public() else AddressType.RANDOM
        )

    def create_disconnected(self, reason: int, conn_handle: int) -> HubMessage:
        """Create a Disconnected message.

        :param reason: Disconnection reason
        :type reason: int
        :param conn_handle: Connection handle
        :type conn_handle: int
        :return: instance of Disconnected
        :rtype: Disconnected
        """
        return BleDomain.bound("disconnected", self.proto_version)(
            reason=reason,
            conn_handle=conn_handle
        )

    def create_desynchronized(self, accesss_address: int) -> HubMessage:
        """Create a Desynchronized message.

        :param access_address: Connection access address
        :type access_address: int
        :return: instance of Desynchronized
        :rtype: Desynchronized
        """
        return BleDomain.bound("desynchronized", self.proto_version)(
            accesss_address=accesss_address
        )

    def create_set_adv_data(self, adv_data: bytes, scan_rsp: bytes = None) -> HubMessage:
        """Create a SetAdvData message.

        :param adv_data: Advertising data
        :type adv_data: bytes
        :param scan_rsp: Scan response data
        :type scan_rsp: bytes, optional
        :return: instance of SetAdvData message
        :rtype: SetAdvData
        """
        message = BleDomain.bound("set_adv_data", self.proto_version)(
            scan_data=adv_data
        )

        # Set scan response data if provided
        if scan_rsp is not None:
            message.scanrsp_data = scan_rsp

        return message

    def create_send_raw_pdu(self, direction: int, pdu: bytes, \
                         crc: int = None, encrypt: bool = False, \
                         access_address: int = None, conn_handle: int = None) -> HubMessage:
        """Create a SendRawPdu message.

        :param direction: PDU direction
        :type direction: int
        :param pdu: PDU to send
        :type pdu: bytes
        :param crc: PDU CRC value
        :type crc: int, optional
        :param conn_handle: Connection handle
        :type conn_handle: int, optional
        :param encrypt: Encrypt PDU before sending
        :type encrypt: bool, optional
        :param access_address: Connection access address
        :type access_address: int, optional
        :return: instance of SendRawPdu message
        :rtype: SendRawPdu
        """
        # Create a SendRawPdu message
        message = BleDomain.bound("send_raw_pdu", self.proto_version)(
            direction=direction,
            pdu=pdu
        )

        # Set optional fields
        if conn_handle is not None:
            message.conn_handle = conn_handle
        if crc is not None:
            message.crc = crc
        if access_address is not None:
            message.access_address= access_address
        if encrypt:
            message.encrypt=True
        else:
            message.encrypt=False

        # Return message
        return message

    def create_send_pdu(self, direction: int, pdu: bytes, conn_handle: int, \
                      encrypt: bool = False) -> HubMessage:
        """Create a SendBlePdu message.

        :param direction: PDU direction
        :type direction: int
        :param pdu: PDU to send
        :type pdu: bytes
        :param conn_handle: Connection handle
        :type conn_handle: int, optional
        :param encrypt: Encrypt PDU before sending
        :type encrypt: bool, optional
        :return: instance of SendRawPdu message
        :rtype: SendBleRawPdu
        """
        return BleDomain.bound("send_pdu", self.proto_version)(
            direction=direction,
            conn_handle=conn_handle,
            pdu=pdu,
            encrypt=encrypt
        )

    def create_adv_pdu_received(self, adv_type: AdvType, rssi: int, bd_address: BDAddress, \
                             adv_data: bytes):
        """Create an AdvPduReceived message

        :param adv_type: Advertisement type
        :type adv_type: AdvType
        :param rssi: Received Signal Strength Indicator
        :type rssi: int
        :param bd_address: Advertiser BD address
        :type bd_address: BDAddress
        :param adv_data: Advertisement data
        :type adv_data: bytes
        :return: instance of AdvPduReceived
        :rtype: AdvPduReceived
        """
        return BleDomain.bound("adv_pdu", self.proto_version)(
            adv_type=adv_type,
            rssi=rssi,
            bd_address=bd_address.value,
            adv_data=adv_data,
            addr_type=AddressType.PUBLIC if bd_address.is_public() else AddressType.RANDOM
        )

    def create_pdu_received(self, direction: int, pdu: bytes, conn_handle: int, \
                          processed: bool = False, decrypted: bool = False) -> HubMessage:
        """Create a PduReceived message

        :param direction: PDU direction
        :type direction: int
        :param pdu: PDU to send
        :type pdu: bytes
        :param conn_handle: Connection handle
        :type conn_handle: int, optional
        :param processed: Set to True if PDU has been processed by firmware
        :type processed: bool
        :param decrypted: Set to True if PDU has been decrypted by firmware
        :type decrypted: bool
        :return: instance of PduReceived
        :rtype: PduReceived
        """
        return BleDomain.bound("pdu", self.proto_version)(
            direction=direction,
            pdu=pdu,
            conn_handle=conn_handle,
            processed=processed,
            decrypted=decrypted
        )

    def create_raw_pdu_received(self, direction: int, pdu: bytes, \
                             access_address: int = None, conn_handle: int = None, \
                             rssi: int = None, timestamp: int = None, \
                             rel_timestamp: int = None, crc: int = None, \
                             crc_validity: bool = None, processed: bool = False, \
                             decrypted: bool = False, channel: int = None) -> HubMessage:
        """Create a RawPduReceived message

        :param direction: PDU direction
        :type direction: int
        :param pdu: PDU to send
        :type pdu: bytes
        :param crc: PDU CRC value
        :type crc: int, optional
        :param conn_handle: Connection handle
        :type conn_handle: int, optional
        :param access_address: Connection access address
        :type access_address: int, optional
        :param rssi: Received Signal Strength Indicator
        :type rssi: int, optional
        :param timestamp: Reception timestamp
        :type timestamp: int, optional
        :param rel_timestamp: Reception timestamp relative to last anchor
        :type rel_timestamp: int, optional
        :param crc_validity: CRC validity status
        :param processed: Set to True if PDU has been processed by firmware
        :type processed: bool
        :param decrypted: Set to True if PDU has been decrypted by firmware
        :type decrypted: bool
        :return: instance of RawPduReceived
        :rtype: RawPduReceived
        """
        # Build message with mandatory fields
        message = BleDomain.bound("raw_pdu", self.proto_version)(
            direction=direction,
            pdu=pdu,
            processed=processed,
            decrypted=decrypted
        )

        # Add optional fields if any
        if rssi is not None:
            message.rssi = rssi
        if access_address is not None:
            message.access_address = access_address
        if conn_handle is not None:
            message.conn_handle = conn_handle
        if timestamp is not None:
            message.timestamp = timestamp
        if rel_timestamp is not None:
            message.relative_timestamp = rel_timestamp
        if crc_validity is not None:
            message.crc_validity = crc_validity
        if crc is not None:
            message.crc = crc
        if channel is not None:
            message.channel = channel

        # Return message
        return message

    def create_injected(self, access_address: int, success: bool, attempts: int) -> HubMessage:
        """Create an Injected message

        :param access_address: Target connection access address
        :type access_address: int
        :param success: Set to True if injection succeeded, False otherwise
        :type success: bool
        :param attempts: Number of attempts
        :type attempts: int
        :return: instance of Injected message
        :rtype: Injected
        """
        return BleDomain.bound("injected", self.proto_version)(
            access_address=access_address,
            success=success,
            injection_attempts=attempts
        )

    def create_hijack_master(self, access_address: int) -> HubMessage:
        """Create a HijackMaster message

        :param access_address: Target access address
        :type access_address: int
        :return: instance of HijackMaster
        :rtype: HijackMaster
        """
        return BleDomain.bound("hijack_master", self.proto_version)(
            access_address=access_address
        )

    def create_hijack_slave(self, access_address: int) -> HubMessage:
        """Create a HijackSlave message

        :param access_address: Target access address
        :type access_address: int
        :return: instance of HijackSlave
        :rtype: HijackSlave
        """
        return BleDomain.bound("hijack_slave", self.proto_version)(
            access_address=access_address
        )

    def create_hijack_both(self, access_address: int) -> HubMessage:
        """Create a HijackBoth message

        :param access_address: Target access address
        :type access_address: int
        :return: instance of HijackBoth
        :rtype: HijackBoth
        """
        return BleDomain.bound("hijack_both", self.proto_version)(
            access_address=access_address
        )

    def create_hijacked(self, access_address: int, success: bool) -> HubMessage:
        """Create an Hijacked message

        :param access_address: Target connection access address
        :type access_address: int
        :param success: Set to True if injection succeeded, False otherwise
        :type success: bool
        :return: instance of Hijacked message
        :rtype: Hijacked
        """
        return BleDomain.bound("hijacked", self.proto_version)(
            access_address=access_address,
            success=success
        )

    def create_prepare_sequence_manual(self, seq_id: int, direction: int, packets: List[bytes]) -> HubMessage:
        """Create a PrepareSequenceManual message

        :param seq_id: Sequence identifier (must be unique)
        :type seq_id: int
        :param direction: Direction of the PDU
        :type direction: Direction
        :param packets: List of PDUs to send
        :type packets: list
        :return: instance of PrepareSequenceManual
        :rtype: PrepareSequenceManual
        """
        # Create our PrepareSequenceManual message
        message = BleDomain.bound("prepare_manual", self.proto_version)(
            sequence_id=seq_id,
            direction=direction
        )

        # Add packets
        for packet in packets:
            message.add_packet(packet)

        # Return message
        return message

    def create_prepare_sequence_conn_evt(self, seq_id: int, direction: int, conn_evt: int, \
                                     packets: List[bytes]) -> HubMessage:
        """Create a PrepareSequenceConnEvt message

        :param seq_id: Sequence identifier (must be unique)
        :type seq_id: int
        :param direction: Direction of the PDU
        :type direction: Direction
        :param conn_evt: target connection event
        :type conn_evt: int
        :param packets: List of PDUs to send
        :type packets: list
        :return: instance of PrepareSequenceConnEvt
        :rtype: PrepareSequenceConnEvt
        """
        message = BleDomain.bound("prepare_connevt", self.proto_version)(
            sequence_id=seq_id,
            direction=direction,
            conn_evt=conn_evt
        )

        # Add packets
        for packet in packets:
            message.add_packet(packet)

        # Return message
        return message

    def create_prepare_sequence_pattern(self, seq_id: int, direction: int, pattern: bytes, \
                                     mask: bytes, offset: int, packets: List[bytes]) -> HubMessage:
        """Create a PrepareSequencePattern message

        :param seq_id: Sequence identifier (must be unique)
        :type seq_id: int
        :param direction: Direction of the PDU
        :type direction: Direction
        :param pattern: target pattern
        :type pattern: bytes
        :param mask: target pattern bitmask
        :type mask: bytes
        :param offset: target pattern offset
        :type offset: int
        :param packets: List of PDUs to send
        :type packets: list
        :return: instance of PrepareSequencePattern
        :rtype: PrepareSequencePattern
        """
        message = BleDomain.bound("prepare_pattern", self.proto_version)(
            sequence_id=seq_id,
            direction=direction,
            pattern=pattern,
            mask=mask,
            offset=offset
        )

        # Add packets
        for packet in packets:
            message.add_packet(packet)

        # Return message
        return message

    def create_triggered(self, seq_id: int) -> HubMessage:
        """Create a Triggered message

        :param seq_id: Sequence identifier triggered
        :type seq_id: int
        :return: instance of Triggered
        :rtype: Triggered
        """
        return BleDomain.bound("triggered", self.proto_version)(
            seq_id=seq_id
        )

    def create_trigger(self, seq_id: int) -> HubMessage:
        """Create a Trigger message

        :param seq_id: Sequence identifier to trigger
        :type seq_id: int
        :return: instance of Trigger
        :rtype: Trigger
        """
        return BleDomain.bound("trigger", self.proto_version)(
            sequence_id=seq_id
        )

    def create_delete_sequence(self, seq_id: int) -> HubMessage:
        """Create a DeleteSequence message

        :param seq_id: Sequence identifier to delete
        :type seq_id: int
        :return: instance of DeleteSequence
        :rtype: DeleteSequence
        """
        return BleDomain.bound("delete_seq", self.proto_version)(
            sequence_id=seq_id
        )

    def create_set_encryption(self, conn_handle: int, ll_key: bytes, ll_iv: bytes, \
                            key: bytes, rand: bytes, ediv: bytes, enabled: bool) -> HubMessage:
        """Create a SetEncryption message

        :param conn_handle: Connection handle
        :type conn_handle: int
        :param ll_key: Link-layer encryption key
        :type ll_key: bytes
        :param ll_iv: Link-layer encryption IV
        :type ll_iv: bytes
        :param key: Encryption key
        :type key: bytes
        :param rand: Encryption random value
        :type rand: bytes
        :param ediv: Encryption diversifier
        :type ediv: bytes
        :param enabled: Enable encryption if set to True, disable it otherwise
        :type enabled: bool
        :return: instance of `SetEncryption`
        """
        return BleDomain.bound("encryption", self.proto_version)(
            conn_handle=conn_handle,
            enabled=enabled,
            ll_key=ll_key,
            ll_iv=ll_iv,
            key=key,
            rand=rand,
            ediv=ediv
        )

from .address import SetBdAddress
from .sniffing import SniffAdv, SniffConnReq, SniffAccessAddress, SniffActiveConn, \
    AccessAddressDiscovered
from .jamming import JamAdv, JamAdvChan, JamConn, ReactiveJam
from .mode import ScanMode, AdvMode, CentralMode, PeriphMode, BleStart, BleStop, \
    SetEncryption
from .pdu import SetAdvData, SendBleRawPdu, SendBlePdu, BleAdvPduReceived, BlePduReceived, \
    BleRawPduReceived, Injected
from .connect import ConnectTo, Disconnect, Connected, Disconnected, Synchronized, \
    Desynchronized
from .hijack import HijackMaster, HijackSlave, HijackBoth, Hijacked
from .triggers import PrepareSequenceManual, PrepareSequenceConnEvt, \
    PrepareSequencePattern, PrepareSequence, Triggered, Trigger, DeleteSequence

__all__ = [
    "AdvType",
    "Direction",
    "AddressType",
    "BleDomain",
    "SetBdAddress",
    "SniffAdv",
    "SniffConnReq",
    "SniffAccessAddress",
    "SniffActiveConn",
    "AccessAddressDiscovered",
    "JamAdv",
    "JamAdvChan",
    "JamConn",
    "ReactiveJam",
    "ScanMode",
    "AdvMode",
    "CentralMode",
    "PeriphMode",
    "SetAdvData",
    "SendBleRawPdu",
    "SendBlePdu",
    "BleAdvPduReceived",
    "BlePduReceived",
    "BleRawPduReceived",
    "ConnectTo",
    "Disconnect",
    "Connected",
    "Disconnected",
    "SetEncryption",
    "BleStart",
    "BleStop",
    "HijackMaster",
    "HijackSlave",
    "HijackBoth",
    "Hijacked",
    "Injected",
    "Synchronized",
    "Desynchronized",
    "PrepareSequenceManual",
    "PrepareSequenceConnEvt",
    "PrepareSequencePattern",
    "PrepareSequence",
    "Triggered",
    "Trigger",
    "DeleteSequence"
]



"""Bluetooth Low Energy domain exceptions
"""

class InvalidSerialPortException(Exception):
    """Exception raised whan an invalid serial port is provided
    """

class InvalidUUIDException(Exception):
    """Exception raised when an invalid UUID is used.
    """


class InvalidHandleValueException(Exception):
    """Exception raised when an invalid handle value is used.
    """


class InvalidAccessAddressException(Exception):
    """Exception raised when an invalid access address is used.
    """


################################
# Connection related exceptions
################################

class PeripheralNotFound(Exception):
    """This exception is raised when a targeted peripheral cannot be found.
    """
    def __init__(self, peripheral=None):
        super().__init__()
        self.peripheral = peripheral

class NotConnected(Exception):
    """This exception is raised when a peripheral is used while not connected.
    """
    def __init__(self, peripheral=None):
        super().__init__()
        self.peripheral = peripheral

class NotSynchronized(Exception):
    """This exception is raised when a connection is used while not synchronized.
    """

    def __repr__(self):
        return "NotSynchronized"

class ConnectionLostException(Exception):
    """This exception is raised when a connection is unexpectedly terminated.
    """
    def __init__(self, connection=None):
        super().__init__()
        self.connection = connection


#######################
# Hook Exceptions
#######################

class HookDontForward(Exception):
    """Raise this exception in a hook to avoid forwarding.
    """

class HookReturnValue(Exception):
    """Raise this exception in a hook to return a specific value to the
    caller.
    """

    def __init__(self, value: bytes):
        super().__init__()
        self.__value = value

    @property
    def value(self) -> bytes:
        """Return value
        """
        return self.__value

class HookReturnGattError(Exception):
    """Raise this exception in a hook to return a GATT error.
    """

    def __init__(self, request, handle, error):
        """Create a HookReturnGattError with request, handle and error code specified.

        :param int request: GATT request
        :param int handle: GATT handle
        :param int error: Error code (from BleAttErrorCode)
        """
        super().__init__()
        self.__request = request
        self.__handle = handle
        self.__error = error

    @property
    def request(self):
        """Corresponding GATT request
        """
        return self.__request

    @property
    def handle(self) -> int:
        """Related attribute GATT handle value
        """
        return self.__handle

    @property
    def error(self) -> int:
        """Error code
        """
        return self.__error


class HookReturnNotFound(Exception):
    """Raise this exception in a hook to indicate the requested object has not
    been found.
    """

class HookReturnAccessDenied(Exception):
    """Raise this exception in a hook to return an error (access not allowed)
    """


class HookReturnAuthentRequired(Exception):
    """Raise this exception in a hook to ask for authentication.
    """


class HookReturnAuthorRequired(Exception):
    """Raise this exception in a hook to notify an Authorization is required
    to access an attribute.
    """

#######################
#Crypto Exceptions
#######################

class MissingCryptographicMaterial(Exception):
    """Raise this exception if a decryption is attempted without having
    cryptographic material available.
    """



""" BLE GATT Attribute Data List and items.
"""
from struct import pack, unpack

from whad.ble.profile import UUID

class GattListItem(object):
    """Template class for a GATT Attribute List Item
    """

    @staticmethod
    def from_bytes(data):
        """Convert a byte array into a list item

        :param bytes data: Data to convert to a list item
        """
        return None

    def to_bytes(self):
        """Serialize list item to bytes
        """
        return b''

    def size(self):
        """Get list item size
        """
        return len(self.to_bytes())


class GattHandleItem(GattListItem):
    """GATT Attribute Data List item used to store start and end handle values.
    """
    def __init__(self, handle, end):
        """
        :param int handle: Attribute handle
        :param int end: Group end
        """
        self.__handle = handle
        self.__end = end

    @property
    def handle(self):
        """Return attribute handle value
        """
        return self.__handle

    @property
    def end(self):
        """Return group end handle value
        """
        return self.__end

    @staticmethod
    def from_bytes(data):
        """Convert bytes to a GattHandleItem object

        :param bytes data: Serialized handle/end pair
        """
        if len(data) == 4:
            handle, end = unpack('<HH', data)
            return GattHandleItem(handle, end)
        else:
            return None

    def to_bytes(self):
        return pack('<HH', self.handle, self.end)

class GattHandleUUIDItem(GattListItem):
    """GATT Attribute Data List item that stores a handle/UUID{16,128} pair.
    """
    def __init__(self, handle, uuid):
        """
        :param int handle: Attribute handle
        :param UUID uuid: Attribute UUID
        """
        self.__handle = handle
        self.__uuid = uuid

    @property
    def handle(self):
        return self.__handle

    @property
    def uuid(self):
        return self.__uuid

    @staticmethod
    def from_bytes(data):
        if len(data)>=4:
            handle = unpack('<H', data[:2])[0]
            uuid = UUID(data[2:])
            return GattHandleUUIDItem(handle, uuid)
        else:
            return None

    def to_bytes(self):
        return pack('<H', self.__handle) + self.__uuid.packed

    def size(self):
        return len(self.to_bytes())

class GattGroupTypeItem(GattListItem):
    """GATT Attribute Data list item that stores group information (start handle, end handle, value).
    """
    def __init__(self, handle, end, value):
        self.__handle = handle
        self.__end = end
        self.__value = value

    @property
    def handle(self):
        return self.__handle

    @property
    def end(self):
        return self.__end

    @property
    def value(self):
        return self.__value

    @staticmethod
    def from_bytes(data):
        if len(data)>4:
            handle, end = unpack('<HH', data[:4])
            value = data[4:]
            return GattGroupTypeItem(handle, end, value)
        return None

    def to_bytes(self):
        return pack('<HH', self.__handle, self.__end) + bytes(self.__value)

class GattAttributeValueItem(GattListItem):
    """GATT Attribute Data list item that stores a handle/value pair.
    """
    def __init__(self, handle, value):
        """
        :param int handle: Attribute handle
        :param bytes value: Attribute value
        """
        self.__handle = handle
        self.__value = value

    @property
    def handle(self):
        return self.__handle

    @property
    def value(self):
        return self.__value

    @staticmethod
    def from_bytes(data):
        if len(data) > 2:
            handle = unpack('<H', data[:2])[0]
            value = data[2:]
            return GattAttributeValueItem(handle, value)
        return None

    def to_bytes(self):
        return pack('<H', self.__handle) + bytes(self.__value)

    def size(self):
        return len(self.to_bytes())

class GattAttributeDataList(object):
    """Generic GATT Attribute Data List class.

    This class can be inherited to provide a list behavior and serialization/deserialization
    functions. A GATT Attribute Data List item class must be provided.
    """
    def __init__(self, item_size):
        self.__item_size = item_size
        self.__items = []

    def __len__(self):
        return len(self.__items)

    def __getitem__(self, idx):
        if idx >= 0 and idx < len(self.__items):
            return self.__items[idx]
        else:
            raise IndexError

    def append(self, item):
        if item.size() == self.__item_size:
            self.__items.append(item)
        else:
            print('wrong size (%d instead of %d)' % (item.size(), self.__item_size))

    def remove(self, item):
        self.__items.remove(item)

    @staticmethod
    def from_bytes(data, item_size, item_clazz):
        """Create GattAttributeDataList from raw data

        :param bytes data: Raw data to parse
        """
        adl = GattAttributeDataList(item_size)
        nb_items = int(len(data)/item_size)
        for i in range(nb_items):
            item = data[i*item_size:(i+1)*item_size]
            adl.append(item_clazz.from_bytes(item))
        return adl

    def to_bytes(self):
        output=b''
        for item in self.__items:
            output += item.to_bytes()
        return output


def gatt_attr_list_iter(item_size, data):
    nb_items = int(len(data)/item_size)
    for i in range(nb_items):
        yield data[i*item_size:(i+1)*item_size]



"""
This module provides the :py:class:`.PeripheralDevice` class used to wrap all GATT operations
for a given connected device:

* discovering services, characteristics and descriptors
* reading a characteristic's value
* writing to a characteristic's value
* subscribing for notifications and indications
* exchanging MTU value with the remote peripheral
"""
import logging
from struct import unpack
from time import sleep
from typing import Iterator, Optional, Union, Type, TypeVar, List

from whad.ble.profile import PrimaryService
from whad.ble.profile.service import Service
from whad.ble.profile.services.defs import ServiceNotFound
from whad.ble.profile.characteristic import (
    Properties, Characteristic, CharacteristicValue,
    Descriptor,
)
from whad.ble.profile import GenericProfile
from whad.ble.profile.attribute import UUID, Attribute, InvalidUUIDException

logger = logging.getLogger(__name__)

PT = TypeVar("PT")

class RemoteAttribute:
    """Remote GATT attribute interface.

    This interface provides the required GATT procedure for any GATT attribute,
    based on an existing GATT layer corresponding to an existing connection already
    established with a remote GATT server.

    Each procedure is started by driving our underlying BLE stack.
    """

    def __init__(self, handle, gatt):
        """Initialize our GATT interface for the specified attribute.

        :param handle: Attribute handle
        :type  handle: int
        :param gatt: GATT layer instance (GATTClient)
        :type gatt: GattClient
        """
        self.__handle = handle
        self.__gatt = gatt

    @property
    def gatt(self):
        """GATT client accessor."""
        return self.__gatt

    def read(self, offset: int = 0) -> bytes:
        """Read the remote attribute using a GATT read or a GATT blob read procedure."""
        if offset == 0:
            return self.__gatt.read(self.__handle)
        return self.read_blob(offset)

    def read_blob(self, offset: int) -> bytes:
        """Read the remote attribute using a GATT blob read procedure."""
        return self.__gatt.read_blob(self.__handle, offset)

    def read_long(self) -> bytes:
        """Read the remote attribute using a combination of classic read
        and blob read requests, depending on the attribute value's length."""
        return self.__gatt.read_long(self.__handle)

    def write(self, value: bytes, without_response: bool = False) -> bool:
        """Write value into the remote attribute using a GATT write precedure."""
        if not without_response:
            result = self.__gatt.write(self.__handle, value)
        else:
            result = self.__gatt.write_command(self.__handle, value)

        # Force result to a boolean value (GATT client's write() method could
        # return None)
        return result == True

    def write_command(self, value: bytes) -> bool:
        """Write value into the remote attribute using a GATT write command procedure."""
        result = self.__gatt.write_command(self.__handle, value)

        # Force result to be a bool
        return result == True

    def write_long(self, value: bytes) -> bool:
        """Write value into the remote attribute using a GATT prepared write procedure."""
        result = self.__gatt.write_long(self.__handle, value)

        # Force result to a bool
        return result == True

class PeripheralCharacteristicDescriptor(Descriptor, RemoteAttribute):
    """Wrapper for a peripheral characteristic descriptor.
    """

    def __init__(self, descriptor, gatt):
        """Initialize a PeripheralCharacteristicDescriptor.

        :param CharacteristicDescriptor descriptor: Descriptor to wrap.
        :param GattClient gatt: GATT client to use for GATT operations (read/write).
        """
        Descriptor.__init__(self, descriptor.uuid, descriptor.handle, descriptor.value,
                                          descriptor.characteristic)
        RemoteAttribute.__init__(self, descriptor.handle, gatt)

    @property
    def value(self) -> bytes:
        """Transparent characteristic read.

        :return bytes: Characteristic value
        """
        value = self.read()

        # Update the underlying attribute value
        if Attribute.value.fset:
            Attribute.value.fset(self, value)
        return value

    @value.setter
    def value(self, value: bytes):
        """Transparent characteristic write.

        :param bytes val: Value to write into this characteristic
        """
        super().write(value)
        # Update the underlying attribute value
        if Attribute.value.fset:
            Attribute.value.fset(self, value)

    @property
    def cached_value(self) -> bytes:
        if Attribute.value.fget:
            return Attribute.value.fget(self)
        return b''

class PeripheralCharacteristicValue(CharacteristicValue, RemoteAttribute):
    """CharacteristicValue wrapper for peripheral devices

    Forward all read/write operations to PeripheralCharacteristic wrapper
    because initially access to characteristic has been implemented there.

    Could be interesting in the future to implement all these operations here
    and to forward from characteristic wrapper to characteristic value wrapper
    because it makes more sense.

    Anyway, this code just works but from an architectural point-of-view is
    a bit crappy.
    """

    def __init__(self, char_value, gatt):
        CharacteristicValue.__init__(self, char_value.uuid, char_value.handle, char_value.value, char_value.characteristic)
        RemoteAttribute.__init__(self, char_value.handle, gatt)

    @property
    def value(self) -> bytes:
        """Transparent characteristic read.

        :return bytes: Characteristic value
        """
        value = super().read()

        # Update model's cached value
        if Attribute.value.fset:
            Attribute.value.fset(self, value)

        return value

    @value.setter
    def value(self, value: bytes):
        """Transparent characteristic write.

        :param bytes val: Value to write into this characteristic
        """
        # Write value
        self.write(value)

        # Update model's cached value
        if Attribute.value.fset:
            Attribute.value.fset(self, value)


class PeripheralCharacteristic(Characteristic, RemoteAttribute):
    """Characteristic wrapper for peripheral devices

    Instruments gatt to read/write a remote characteristic.
    """
    def __init__(self, characteristic, gatt):
        # Populate this characteristic attribute
        Characteristic.__init__(self, characteristic.uuid, characteristic.handle,
                         characteristic.end_handle, characteristic.value,
                         characteristic.properties, characteristic.security)

        # Wrap descriptors and add them to our list of descriptors
        for desc in characteristic.descriptors():
            self.add_descriptor(PeripheralCharacteristicDescriptor(desc, gatt))

        # Initialize the remote attribute interface
        RemoteAttribute.__init__(self,characteristic.value_handle, gatt)

    def get_descriptor(self, desc_type: Union[UUID, Type[Descriptor]]):
        """Retrieve a specific descriptor from those associated with this characteristic."""
        result = super().get_descriptor(desc_type)
        # Not found? return None.
        if result is None:
            return None

        # Wrap descriptor if needed.
        if not isinstance(result, PeripheralCharacteristicDescriptor):
            return PeripheralCharacteristicDescriptor(result, self.gatt)
        return result

    @property
    def value(self) -> bytes:
        """Characteristic's value"""
        # Read value from characteristic
        value = super().read()

        # Update the underlying attribute value
        if Attribute.value.fset:
            Attribute.value.fset(self.value_attr, value)
        return value

    @value.setter
    def value(self, value: bytes):
        # Write new value into characteristic
        self.write(value)

        # Update model's cached value
        if Attribute.value.fset:
            Attribute.value.fset(self.value_attr, value)

    @property
    def cached_value(self) -> bytes:
        if Attribute.value.fget:
            return Attribute.value.fget(self.value_attr)
        return b''

    def read(self, offset: int = 0, long: bool = False) -> bytes:
        """Read characteristic value.

        :param offset: If specified, start reading at this offset.
        :type  offset: int
        :param long: If enabled, perform a long read.
        :ttype long: bool
        :return: Content of the characterstic's value
        :rtype:  bytes
        """
        if long:
            return super().read_long()
        else:
            return super().read(offset=offset)

    def write(self, value: bytes, without_response: bool = False) -> bool:
        """Set characteristic value

        If characteristic is only writeable without response, use a write command
        rather than a write request. Otherwise, use a write request. If a characteristic
        has both write and write without response properties, `without_response` must be
        set to True to use a write command.

        :param value: Value to write into the characteristic
        :type value: bytes
        :param without_response: Send a GATT write command instead of a GATT write
                                 if set to `True`
        :return: `True` on successful write, `False` otherwise
        :rtype: bool
        """
        # If characteristic is only writeable without response, force without_response to True.
        access_mask = Properties.WRITE_WITHOUT_RESPONSE | Properties.WRITE
        if (self.properties & access_mask) == Properties.WRITE_WITHOUT_RESPONSE:
            without_response = True

        return super().write(value, without_response)

    def subscribe(self, notification=False, indication=False, callback=None):
        """Subscribe for notification/indication.

        :param bool notification: If set, subscribe for notification
        :param bool indication: If set, subscribe for indication (cannot be used
                                when notification is set)
        :param callable callback: Callback function to be called on
                                  indication/notification event
        :return bool: True if subscription has successfully been performed, False otherwise.
        """
        # wrap our callback to provide more details about the concerned
        # characteristic
        def wrapped_cb(_, value, indication=False):
            if callback is not None:
                callback(
                    self,
                    value,
                    indication=indication
                )

        if notification:
            # Look for CCCD
            desc = self.get_descriptor(UUID(0x2902))
            if desc is not None:
                # Register our callback
                if callback is not None:
                    self.gatt.register_notification_callback(
                        self.value_handle,
                        wrapped_cb
                    )

                # Enable notification
                desc.write(bytes([0x01, 0x00]))

                return True

            # No CCCD, cannot subscribe
            logger.debug("No CCC descriptor, cannot subscribe to charac. %s", self.uuid)
            return False

        if indication:
            # Look for CCCD
            desc = self.get_descriptor(UUID(0x2902))
            if desc is not None:
                # Register our callback
                if callback is not None:
                    self.gatt.register_notification_callback(
                        self.value_handle,
                        wrapped_cb
                    )

                # Enable indication
                desc.write(bytes([0x02, 0x00]))

                return True

            # No CCCD, cannot subscribe for indications
            logger.debug("No CCC descriptor, cannot subscribe to charac. %s", self.uuid)
            return False

        # No indication or notification subscription required
        return False

    def unsubscribe(self):
        """Unsubscribe from this characteristic.
        """
        # Look for CCCD
        desc = self.get_descriptor(UUID(0x2902))

        if desc is not None:
            # Disable notification/indication
            desc.write(bytes([0x00, 0x00]))

            # Unregister our callback
            self.gatt.unregister_notification_callback(
                self.value_handle
            )
            return True

        # No descriptor found, cannot unsubscribe
        logger.debug("No CCC descriptor, cannot unsubscribe from charac. %s", self.uuid)
        return False


class PeripheralService(Service):
    """Service wrapper for peripheral devices
    """

    def __init__(self, service, gatt):
        """Initialize a peripheral service from discovered GATT service."""
        self.__gatt = gatt
        super().__init__(service.type_uuid, service.uuid, service.handle, service.end_handle)

        # Copy characteristics and wrap each of them into a `PeripheralCharacteristic` object.
        for charac in service.characteristics():
            self.add_characteristic(PeripheralCharacteristic(charac, self.__gatt))

        # Copy included services
        for inc_service in service.included_services():
            self.add_included_service(inc_service)

    def __iter__(self) -> Iterator[UUID]:
        """Iterator over the discovered characteristics."""
        for charac in self.characteristics():
            yield charac.uuid

    def __getitem__(self, key) -> PeripheralCharacteristic:
        """
        Retrieve a :class:`whad.ble.profile.device.PeripheralCharacteristic` object from
        its corresponding UUID.

        This special method simplifies operations based on this :class:`whad.ble.profile.device.PeripheralService`
        class, like checking the presence of a characteristing in the discovered service or simply getting an object
        representing this characteristic.

        To check if a specific characteristic with UUID `2A00` exists, one may simply use Python's `in` operator:

        ```python
        # From the 16-bit integer value corresponding to the characteristic UUID:
        if 0x2A00 in service:
            print("Characteristic with UUID 2A00 does exist.")

        # Or a string containing the UUID value
        if '2A00' in service:
            print("Characteristic with UUID 2A00 does exist.")

        # Or simply with an instance of UUID
        if UUID('2A00') in service:
            print("Characteristic with UUID 2A00 does exist.")

        ```

        To access a specific characteristic and retrieve the corresponding
        :class:`whad.ble.profile.device.PeripheralCharacteristic` object:

        ```python
        char = service[UUID('1800')]
        ```
        """
        # Key must be a valid UUID
        try:
            # Check service UUID
            if not isinstance(key, UUID):
                char_uuid = UUID(key)
            else:
                char_uuid = key

            # Retrieve object, if it does exist.
            char = self.char(char_uuid)
            if char is None:
                raise IndexError()

            # Found
            return char
        except InvalidUUIDException as uuid_err:
            raise IndexError() from uuid_err

    def read_characteristic_by_uuid(self, uuid) -> Optional[bytes]:
        """Read a characteristic belonging to this service identified by its UUID, using a GATT ReadByType
        procedure as defined in the specification (Vol 3, Part G, Section 4.8.2).

        This method can be called at any time, even if the target device's attributes have not been discovered
        yet by calling the :py:meth:`~.discover` method.

        :param uuid: Characteristic UUID
        :type  uuid: UUID
        :return bytes: Characteristic value
        """
        values = self.__gatt.read_characteristic_by_uuid(
            uuid,
            self.handle,
            self.end_handle
        )
        if len(values) > 0:
            return values[0].value
        else:
            return None

    def char(self, uuid: Union[str, UUID]) -> Optional[PeripheralCharacteristic]:
        """Look for a specific characteristic belonging to this service, identified by its UUID.

        :param uuid: Characteristic UUID
        :type  uuid: UUID, str
        :return: Found characteristic if any, `None` otherwise.
        :rtype: PeripheralCharacteristic
        :raise: InvalidUUIDException
        """
        # Convert to UUID if str is provided
        if isinstance(uuid, str):
            uuid = UUID(uuid)

        # Search for matching characteristic
        for charac in self.characteristics():
            if charac.uuid == uuid:
                return charac

        # Not found
        return None


    def get_characteristic(self, uuid: UUID) -> Optional[PeripheralCharacteristic]:
        """Look for a specific characteristic belonging to this service, identified by its UUID.

        :param uuid: Characteristic UUID
        :type  uuid: UUID
        :return: Found characteristic if any, `None` otherwise.
        :rtype: PeripheralCharacteristic

        .. deprecated:: 1.3.0
            This method has been superseeded by :py:meth:`~.char` starting from version 1.3.0,
            in a effort to make WHAD's API simpler and easier to use.
        """
        return self.char(uuid)


class PeripheralDevice(GenericProfile):
    """GATT client wrapper representing a remote device.

    This class is used to wrap a device model used in a gatt client
    in order to provide easy-to-use methods to access its services,
    characteristics and descriptors.
    """

    def __init__(self,  central, gatt_client, conn_handle, from_json=None):
        """Create a peripheral device from a Central and a GATT client.

        :param  central:        Central instance used to connect to a target device.
        :type   central:        :class:`whad.ble.connector.central.Central`
        :param  gatt_client:    GATT client connected to a target device.
        :type   gatt_client:    :class:`whad.ble.stack.gatt.GattClient`
        :param  conn_handle:    Current connection handle.
        :type   conn_handle:    int
        :param  from_json:      GATT profile (JSON) to be used when instanciating
                                the underlying GattProfile.
        :type   from_json:      str, optional
        """
        self.__gatt = gatt_client
        self.__smp = gatt_client.smp
        self.__ll = gatt_client.get_layer('ll')
        self.__conn_handle = conn_handle
        self.__central = central
        self.__disconnect_cb = None
        super().__init__(from_json=from_json)


    @property
    def conn_handle(self) -> int:
        """Current connection handle.
        """
        return self.__conn_handle

    def __iter__(self) -> Iterator[UUID]:
        """Iterator for services's UUIDs."""
        for service in self.services():
            yield service.uuid

    def __getitem__(self, key):
        """
        Retrieve a :class:`whad.ble.profile.device.PeripheralService` object from
        its corresponding UUID.

        This special method simplifies operations based on this :class:`whad.ble.profile.device.PeripheralDevice`
        class, like checking the presence of a service in the discovered atrributes or simply getting an object
        representing this service.

        To check if a specific service with UUID `1800` exists, one may simply use Python's `in` operator:

        ```python
        # From the 16-bit integer value corresponding to the service UUID:
        if 0x1800 in device:
            print("Service with UUID 1800 does exist.")

        # Or a string containing the UUID value
        if '1800' in device:
            print("Service with UUID 1800 does exist.")

        # Or simply with an instance of UUID
        if UUID('1800') in device:
            print("Service with UUID 1800 does exist.")

        ```

        To access a specific service and retrieve the corresponding :class:`whad.ble.profile.device.PeripheralService`
        object:

        ```python
        service = device[UUID('1800')]
        ```
        """
        # Key must be a valid UUID
        try:
            # Check service UUID
            if not isinstance(key, UUID):
                service_uuid = UUID(key)
            else:
                service_uuid = key

            # Retrieve object, if it does exist.
            service = self.service(service_uuid)
            if service is None:
                raise IndexError()

            # Found
            return service
        except InvalidUUIDException as uuid_err:
            raise IndexError() from uuid_err


    def start_encryption(self):
        """Start encryption procedure for BLE peripheral
        """
        security_database = self.__smp.security_database


        crypto_material = security_database.get(address=self.__central.target_peer)
        conn_handle = self.__smp.get_layer('l2cap').state.conn_handle
        if crypto_material is not None and crypto_material.has_ltk():
            self.__ll.start_encryption(
                conn_handle,
                unpack('>Q', crypto_material.ltk.rand)[0],
                crypto_material.ltk.ediv
            )

    def pairing(self, pairing=None):
        """Trigger a pairing according to provided parameters.
        Default parameters will be used if pairing parameter is None.
        """
        if not self.__smp.initiate_pairing(parameters=pairing):
            return False

        while not self.__smp.is_pairing_done():
            sleep(0.1)
            if self.__smp.is_pairing_failed():
                return False

        self.__smp.reset_state()
        return True

    def set_disconnect_cb(self, callback):
        """Set disconnection callback.

        :param callback:    Callback function to call on disconnection.
        :type   callback:   callable
        """
        self.__disconnect_cb = callback


    def get_mtu(self) -> int:
        """Retrieve the current host's connection MTU.

        :return: WHAD device MTU
        :rtype: int
        """
        return self.__gatt.att.get_client_mtu()

    def set_mtu(self, mtu: int):
        """Update connection MTU.

        :param  mtu:    ATT MTU to use for this connection.
        :type   mtu:    int
        :return:        Remote device MTU.
        :rtype: int
        """
        return self.__gatt.set_mtu(mtu)


    def disconnect(self):
        """Terminate the connection to this device
        """
        # Ask associated central to disconnect this peripheral device.
        self.__central.disconnect(self.__conn_handle)


    def discover(self, include_values: bool = False):
        """Discovers services, characteristics and descriptors.

        This method must be called before accessing any service or characteristic,
        as it is required to retrieve the corresponding GATT handles.
        """
        # Discover
        self.__gatt.discover(save_values=include_values)

    def discover_primary_services(self, start: int = 1):
        """Discovers primary services only, starting from handle `start`.
        Discovered services are added to the current GATT profile and can be listed
        with :py:meth:`.services()`.

        :param start: Discover services with handle in range [`start`, 0xFFFF]
        :type  start: int
        """
        self.__gatt.discover_primary_services(start)

    def discovery_service_characteristics(self, service: Service, values: bool = False,
                                          start: int = 1) -> Iterator[Characteristic]:
        """Discovers characteristics that belong to a specific service.

        :param service: Service to discover characteristics from
        :type  service: Service
        :param values: When set to `True`, characteristics' value will be read and loaded into their
                       corresponding CharacteristicValue's value
        :type  values: bool, optional
        :param  start: Handle value used as the starting value for this discovery procedure
        :type   start: int, optional
        :return: Iterator over discovered characteristics
        :rtype: Iterator[Characteristic]
        """
        yield from self.__gatt.discovery_service_characteristics(service, save_values=values, start=start)

    def find_service_by_uuid(self, uuid: UUID) -> Optional[Service]:
        """Find service by its UUID

        :param  uuid:   Characteristic UUID
        :type   uuid:   :class:`whad.ble.profile.attribute.UUID`
        :return:        PeripheralService: An instance of PeripheralService if
                        service has been found, None otherwise.
        :rtype: :class:`whad.ble.profile.device.PeripheralService`
        """
        service = self.__gatt.discover_primary_service_by_uuid(uuid)
        if service is not None:
            return PeripheralService(
                service,
                self.__gatt
            )

        # Not found
        return None

    def char(self, uuid: Union[str, UUID], service: Optional[Union[str, UUID]] = None) -> Optional[PeripheralCharacteristic]:
        """Retrieve a characteristic by its UUID. If more than one characteristic is found,
        returns the first match.

        Search can be narrowed to a specific service if the `service` parameter is set.

        :param uuid: Characteristic's UUID
        :type  uuid: :class:`whad.ble.profile.attribute.UUID`, str
        :type  uuid: :class:`str`
        :param service: Service's UUID
        :type  service: :class:`whad.ble.profile.attribute.UUID`, :class:`str`, optional
        :return:     First matching characteristic, `None` if not found
        :rtype:      PeripheralCharacteristic
        :raise:      InvalidUUIDException
        """
        # If UUID is a string, convert to the corresponding UUID object
        if isinstance(uuid, str):
            uuid = UUID(uuid)

        # Search for characteristic
        if service is None:
            for s in self.services():
                for charac in s.characteristics():
                    if charac.uuid == uuid:
                        return charac
        else:
            # Convert service UUID if provided as a string
            if isinstance(service, str):
                service_uuid = UUID(service)
            else:
                service_uuid = service

            s = self.service(service_uuid)
            if s is not None:
                return s.char(uuid)

        # Not found
        return None


    def find_characteristics_by_uuid(self, uuid: UUID) -> List[PeripheralCharacteristic]:
        """Find characteristic by its UUID

        :param  uuid:   Characteristic UUID
        :type   uuid:   :class:`whad.ble.profile.attribute.UUID`
        :return:        PeripheralCharacteristic: An instance of PeripheralCharacteristic
                        if characteristic has been found, None otherwise.
        :rtype: :class:`whad.ble.profile.device.PeripheralCharacteristic`
        """
        characs = self.__gatt.find_characteristics_by_uuid(uuid)
        return [PeripheralCharacteristic(charac, self.__gatt) for charac in characs]

    def find_object_by_handle(self, handle) -> Optional[Attribute]:
        """Find an existing object (service, attribute, descriptor) based on its handle,
        it known from the underlying GenericProfile.

        :param  handle: Object handle
        :type   handle: int
        :return:        Characteristic, characteristic value or service
        :rtype:         :class:`whad.ble.profile.device.PeripheralCharacteristic`,
                        :class:`whad.ble.profile.device.PeripheralCharacteristicValue`,
                        :class:`whad.ble.profile.device.PeripheralService`
        """
        # Search for object
        obj = super().find_object_by_handle(handle)

        # If object has been found, make sure we wrap it in the corresponding class
        # to allow user to read and write from/into this attribute over the existing
        # connection.

        if isinstance(obj, Characteristic):
            # Wrap characteristic if required
            if not isinstance(obj, PeripheralCharacteristic):
                return PeripheralCharacteristic(obj, self.__gatt)
            return obj

        if isinstance(obj, Service):
            # Wrap service if required
            if not isinstance(obj, PeripheralService):
                return PeripheralService(obj, self.__gatt)
            return obj

        if isinstance(obj, CharacteristicValue):
            # Wrap characteristic value if required
            if not isinstance(obj, PeripheralCharacteristicValue):
                return PeripheralCharacteristicValue(
                    obj,
                    self.__gatt
                )
            return obj

        if isinstance(obj, Descriptor):
            # wrap descriptor if required
            if not isinstance(obj, PeripheralCharacteristicDescriptor):
                return PeripheralCharacteristicDescriptor(
                    obj,
                    self.__gatt
                )
            return obj

        # Not found
        return None

    def get_characteristic(self, service_uuid: UUID, charac_uuid: UUID):
        """Get a PeripheralCharacteristic object representing a characteristic
        defined by the given service UUID and characteristic UUID.

        :param  service_uuid:   Service UUID
        :type   service_uuid:   :class:`whad.ble.profile.attribute.UUID`
        :param  charac_uuid:    Characteristic UUID
        :type   charac_uuid:    :class:`whad.ble.profile.attribute.UUID`
        :return:                PeripheralCharacteristic object on success, None if not found.
        :rtype: :class:`whad.ble.profile.device.PeripheralCharacteristic`
        """
        service = self.get_service(service_uuid)
        if service is not None:
            return service.get_characteristic(charac_uuid)
        return None

    def has(self, interface: Type[PT]) -> bool:
        """Check if device exposes a specific service interface."""
        if issubclass(interface, PrimaryService):
            service_model = interface()

            # Raise an exception if service's UUID is undefined
            if service_model.uuid is None:
                return False

            # Check if we have such service
            remote_serv = self.service(service_model.uuid)
            if remote_serv is None:
                return False

            # Check if we have all the required characteristics
            for char in service_model.characteristics():
                # Make sure the service's characteristic has an UUID
                if char.uuid is None:
                    return False

                # Search for mandatory characteristics
                if remote_serv.char(char.uuid) is None:
                    if char.required:
                        return False

            # success if we loaded at least one service
            return True

        # Failure.
        return False


    def query(self, interface: Type[PT]) -> PT:
        """Dynamically load a pluggable service into this device definition."""
        if issubclass(interface, PrimaryService):
            # Create an instance of our associated primary service class. We will later
            # inject our own characteristics into it.
            service = interface()
            if service.uuid is not None:
                # Retrieve the remote service based on this UUID
                remote_service = self.service(service.uuid)
                if remote_service is not None:

                    # Loop on characteristics defined in this service,
                    # search for them in our attribute DB and inject
                    # our wrapped objects into our remote service with
                    # the corresponding aliases
                    for char in service.characteristics():
                        # Make sure the service's characteristic has an UUID
                        if char.uuid is None:
                            raise ServiceNotFound()

                        # Query the corresponding characteristic object
                        # and inject it into our pluggable instance.
                        remote_char = remote_service.char(char.uuid)
                        if remote_char is not None and char.alias:
                            setattr(service, char.alias, remote_char)
                        elif char.required:
                            raise ServiceNotFound()
                        elif char.alias:
                            # Inject the characteristic but set it to None (undefined)
                            setattr(service, char.alias, None)

                    # Return the populated service interface instance
                    return service

        # Pluggable cannot be loaded.
        raise ServiceNotFound()

    def service(self, uuid: Union[str, UUID]):
        """Retrieve a PeripheralService object given its UUID.

        :param  uuid:       Service UUID
        :type   uuid:       :class:`whad.ble.profile.attribute.UUID`
        :type   uuid:       str
        :return:            Corresponding PeripheralService object if found, None otherwise.
        :rtype: :class:`whad.ble.profile.device.PeripheralService`
        :raise: InvalidUUIDException
        """
        # If a string is provided as UUID, convert it to the corresponding
        # UUID object. This could raise an InvalidUUIDException.
        if isinstance(uuid, str):
            uuid = UUID(uuid)

        # Search for a service matching the given UUID
        for service in self.services():
            if service.uuid == uuid:
                if not isinstance(service, PeripheralService):
                    return PeripheralService(service, self.__gatt)
                return service
        return None

    def get_service(self, uuid):
        """Retrieve a PeripheralService object given its UUID.

        :param  uuid:       Service UUID
        :type   uuid:       :class:`whad.ble.profile.attribute.UUID`
        :return:            Corresponding PeripheralService object if found, None otherwise.
        :rtype: :class:`whad.ble.profile.device.PeripheralService`

        .. deprecated:: 1.3.0
            The new :meth:`~whad.ble.profule.device.PeripheralDevice.service` method shall be used to
            retrieve a :class:`~whad.ble.profile.device.PeripheralService` object representing a service
            identified by a given UUID.
        """
        return self.service(uuid)

    def write(self, handle, value):
        """Perform a write operation on an attribute based on its handle.

        This method allows to interact with characteristics and descriptors without
        having performing a GATT services and characteristics discovery. One just need
        to specify the handle corresponding to a characteristic value or descriptor and the
        value to write to it, and our GATT stack will handle it.

        Note that there is absolutely no check on corresponding characteristic permissions
        (meaning you can try to write on a read-only characteristic value) and that this
        method may raise exceptions due to potential GATT errors the remote device may
        return.

        :param  handle: Characteristic or descriptor handle to write.
        :type   handle: int
        :param  value:  Bytes to write into this characteristic.
        :type   value:  bytes
        """
        return self.__gatt.write(handle, value)


    def write_command(self, handle, value):
        """Perform a write command operation (no write response will be sent) on
        an attribute based on its handle.

        This method allows to interact with characteristics and descriptors without
        having performing a GATT services and characteristics discovery. One just need
        to specify the handle corresponding to a characteristic value or descriptor and the
        value to write to it, and our GATT stack will handle it.

        Note that there is absolutely no check on corresponding characteristic permissions
        (meaning you can try to write on a read-only characteristic value) and that this
        method may raise exceptions due to potential GATT errors the remote device may
        return.

        :param  handle: Characteristic or descriptor handle to write.
        :type   handle: int
        :param  value:  Bytes to write into this characteristic.
        :type   value:  bytes
        """
        return self.__gatt.write_command(handle, value)

    def read(self, handle, offset=None, long=False):
        """Perform a read operation on an attribute based on its handle.

        This method allows to interact with characteristics and descriptors without
        having performing a GATT services and characteristics discovery. One just need
        to specify the handle corresponding to a characteristic value or descriptor
        and our GATT stack will handle it.

        Note that there is absolutely no check on corresponding characteristic permissions
        (meaning you can try to read from a write-only characteristic value) and that this
        method may raise exceptions due to potential GATT errors the remote device may
        return.

        :param  handle: Characteristic or descriptor handle.
        :type   handle: int
        :param  offset: Offset applied when reading data from characteristic or
                        descriptor (default: 0).
        :type   offset: int, optional
        :param  long:   use GATT long read procedure if set to True (default: False)
        :type   long:   bool, optional
        :return:        Content of the characteristic or descriptor.
        :rtype: bytes
        """
        if not long:
            if offset is None:
                return self.__gatt.read(handle)

            # Use provided offset
            return self.__gatt.read_blob(handle, offset=offset)

        # Read long
        return self.__gatt.read_long(handle)

    def services(self) -> Iterator[PeripheralService]:
        """Iterate over the device's GATT services.

        :return: An iterator that can be used to iterate over services.
        :rtype: Iterator
        """
        for service in super().services():
            if not isinstance(service, PeripheralService):
                yield PeripheralService(
                    service,
                    self.__gatt
                )
            else:
                yield service

    def on_disconnect(self, conn_handle):
        """Disconnection callback

        :param  conn_handle:    Connection handle
        :type   conn_handle:    int
        """
        logger.debug("PeripheralDevice has disconnected (conn. handle: %d)", conn_handle)
        if self.__disconnect_cb is not None:
            self.__disconnect_cb()

    def on_mtu_changed(self, mtu: int):
        """MTU change callback

        :param mtu: New MTU value
        :type  mtu: int
        """
        logger.debug("PeripheralDevice: MTU has been changed to %d", mtu)



"""WHAD Protocol ESB mode messages abstraction layer.
"""
from whad.protocol.whad_pb2 import Message
from whad.protocol.esb.esb_pb2 import StartCmd, StopCmd
from whad.hub.events import JammedEvt
from ..message import pb_bind, PbFieldBytes, PbFieldBool, PbFieldInt, PbMessageWrapper
from . import EsbDomain

@pb_bind(EsbDomain, 'sniff', 1)
class SniffMode(PbMessageWrapper):
    """ESB SniffMode message
    """

    channel = PbFieldInt('esb.sniff.channel')
    address = PbFieldBytes('esb.sniff.address')
    show_acks = PbFieldBool('esb.sniff.show_acknowledgements')

@pb_bind(EsbDomain, 'jam', 1)
class JamMode(PbMessageWrapper):
    """ESB JamMode message
    """

    channel = PbFieldInt('esb.jam.channel')

@pb_bind(EsbDomain, 'jammed', 1)
class Jammed(PbMessageWrapper):
    """ESB Jammed notification message
    """

    timestamp = PbFieldInt('esb.jammed.timestamp')

    def to_event(self):
        """Convert this message into a WHAD event.
        """
        return JammedEvt(
            timestamp=self.timestamp
        )
    
    @staticmethod
    def from_event(event):
        """Create a Jammed message from event.
        """
        return Jammed(timestamp=event.timestamp)

@pb_bind(EsbDomain, 'start', 1)
class EsbStart(PbMessageWrapper):
    """ESB EsbStart message
    """

    def __init__(self, message: Message = None):
        super().__init__(message=message)
        self.message.esb.start.CopyFrom(StartCmd())

@pb_bind(EsbDomain, 'stop', 1)
class EsbStop(PbMessageWrapper):
    """ESB EsbStop message
    """

    def __init__(self, message: Message = None):
        super().__init__(message=message)
        self.message.esb.stop.CopyFrom(StopCmd())

@pb_bind(EsbDomain, 'prx', 1)
class PrxMode(PbMessageWrapper):
    """ESB PrxMode message
    """

    channel = PbFieldInt('esb.prx.channel')

@pb_bind(EsbDomain, 'ptx', 1)
class PtxMode(PbMessageWrapper):
    """ESB PtxMode message
    """

    channel = PbFieldInt('esb.ptx.channel')


"""Bluetooth Low Energy sniffing classes used by `wsniff` for generic
sniffing.
"""
from dataclasses import dataclass, field

from whad.ble.exceptions import InvalidAccessAddressException
from whad.ble.utils.phy import is_access_address_valid
from whad.common.sniffing import SniffingEvent
from whad.hub.ble import ChannelMap

@dataclass
class SynchronizedConnection:
    """Synchronized connection data class.
    """
    access_address : int = None
    crc_init : int = None
    hop_interval : int = None
    hop_increment : int = None
    channel_map : ChannelMap = None

class ConnectionConfiguration(SynchronizedConnection):
    """
    Configuration for sniffing an existing Bluetooth Low Energy.

    :param access_address: indicate access address of the targeted connection (aa)
    :param crc_init: indicate CRCInit of the targeted connection (crc)
    :param hop_interval: indicate Hop Interval of the targeted connection (int)
    :param hop_increment: indicate Hop Increment of the targeted connection (inc)
    :param channel_map: indicate Channel Map of the targeted connection (chm)
    """

class SynchronizationEvent(SniffingEvent):
    """Synchronization event
    """

    def __init__(self, connection):
        super().__init__("Connection synchronized")
        self.synchronized_connection = connection

    @property
    def message(self):
        """Readable representation of this event
        """
        return (
            f"access_address=0x{self.synchronized_connection.access_address:08x}, "
            f"crc_init=0x{self.synchronized_connection.crc_init:06x}, "
            f"hop_interval={self.synchronized_connection.hop_interval} "
            f"({self.synchronized_connection.hop_interval*1250} us), "
            f"hop_increment={self.synchronized_connection.hop_increment}, "
            f"channel_map=0x{self.synchronized_connection.channel_map.value.hex()}"
        )


class DesynchronizationEvent(SniffingEvent):
    """Event indicating a desynchronization
    """
    def __init__(self):
        super().__init__("Connection desynchronized")



class KeyExtractedEvent(SniffingEvent):
    """Event indicating that a key has been extracted from pairing
    """
    def __init__(self, key):
        super().__init__("Key extracted")
        self.key = key

    @property
    def message(self):
        return f"key={self.key.hex()}"

@dataclass
class SnifferConfiguration:
    """
    Configuration for the Bluetooth Low Energy sniffer.

    :param show_advertisements: enable advertisement sniffing (a)
    :param follow_connection: enable new connection sniffing (f)
    :param show_empty_packets: display empty packets during connection (e)
    :param access_addresses_discovery: discover access addresses of existing connections
    :param pairing: perform attack on legacy pairing (p)
    :param active_connection: enable and configure existing connection sniffing
    :param channel: select the channel to sniff (c)
    :param filter: display only the packets matching the filter BD address (m)
    :param decrypt: indicate if decryption is enabled (d)
    :param keys: provide decryption keys (k)
    """
    show_advertisements : bool = True
    follow_connection : bool = False
    show_empty_packets : bool = False
    access_addresses_discovery : bool = False
    pairing : bool = False
    active_connection : ConnectionConfiguration = None
    channel : int = 37
    filter : str = "FF:FF:FF:FF:FF:FF"
    decrypt : bool = False
    keys : list = field(default_factory=lambda: [])

class AccessAddress:
    """Bluetooth connection Access Address.
    """

    def __init__(self, access_address, timestamp=None, rssi=None):
        if not is_access_address_valid(access_address):
            raise InvalidAccessAddressException()

        self.__access_address = access_address
        self.__timestamp = timestamp
        self.__rssi = rssi
        self.__count = 1

    def __int__(self) -> int:
        """Integer conversion
        """
        return self.__access_address

    def __eq__(self, other) -> bool:
        """Compare two access addresses.
        """
        return int(other) == int(self)

    def update(self, timestamp=None, rssi=None):
        """Update access address RSSI and timestamp, and
        increment the internal counter for statistics.
        """
        self.__count += 1
        if timestamp is not None:
            self.__timestamp = timestamp
        if rssi is not None:
            self.__rssi = rssi

    @property
    def last_timestamp(self) -> int:
        """Last seen timestamp
        """
        return self.__timestamp

    @property
    def last_rssi(self) -> int:
        """Last RSSI
        """
        return self.__rssi

    @property
    def count(self) -> int:
        """Current count
        """
        return self.__count

    def __repr__(self) -> str:
        """String representation
        """
        rssi = self.__rssi if self.__rssi is not None else ""
        timestamp = self.__timestamp if self.__timestamp is not None else ""
        printable_string =  (
            f"0x{self.__access_address:08x} (seen {self.count} times), "
            f"rssi = {rssi}, last_timestamp = {timestamp})"
        )
        return printable_string



from whad.dot15d4.stack.database import Dot15d4Database

class NWKIB(Dot15d4Database):
    """
    RF4CE NWKIB Database of attributes.
    """

    def reset(self):
        """
        Reset the NWKIB database to its default value.
        """
        self.nwkActivePeriod = 0x00041a
        self.nwkBaseChannel = 15
        self.nwkDiscoveryLQIThreshold = 0xFF
        self.nwkDiscoveryRepetitionInterval = 0x0030d4
        self.nwkDutyCycle = 0x000000
        self.nwkFrameCounter = 0x00000001
        self.nwkIndicateDiscoveryRequests = False
        self.nwkInPowerSave = False
        self.nwkPairingTable = []
        self.nwkMaxDiscoveryRepetitions = 1
        self.nwkMaxFirstAttemptCSMABackoffs = 4
        self.nwkMaxFirstAttemptFrameRetries = 3
        self.nwkMaxReportedNodeDescriptors = 3
        self.nwkResponseWaitTime = 0x00186a
        self.nwkScanDuration = 6
        self.nwkUserString = "Telink"
        self.nwkVendorString = "TL"
        self.nwkVendorIdentifier = 4417

        self.nwkcNodeCapabilities = (
            (0 << 3) | # channel_normalization_capable
            (1 << 2) | # security_capable
            (1 << 1) | # power_source
            (1) # node_type
        )



from whad.unifying.connector import Unifying
from whad.esb.stack import ESBStack
from whad.unifying.stack import UnifyingApplicativeLayer, UnifyingRole
from whad.exceptions import UnsupportedCapability


class Dongle(Unifying):
    """
    Logitech Unifying Dongle interface for compatible WHAD device.
    """
    def __init__(self, device, **kwargs):
        super().__init__(device)

        self.__channel = 5
        self.__address = "ca:e9:06:ec:a4"
        self.__started = False
        ESBStack.add(UnifyingApplicativeLayer)
        self.__stack = ESBStack(self)

        for name, callback in kwargs.items():
            if name.startswith("on_"):
                self.__stack.app.callbacks[name] = callback

        # Check if device can choose its own address
        if not self.can_set_node_address():
            raise UnsupportedCapability("SetNodeAddress")

        # Check if device can perform dongle simulation
        if not self.can_be_dongle():
            raise UnsupportedCapability("DongleSimulation")

        self._enable_role()


    def _enable_role(self):
        if self.__started:
            super().stop()
        self.set_node_address(self.__address)
        self.enable_dongle_mode(channel=self.__channel)
        self.__stack.app.role = UnifyingRole.DONGLE
        if self.__started:
            super().start()

    @property
    def channel(self):
        return self.__channel

    @channel.setter
    def channel(self, channel=5):
        self.__channel = channel
        self._enable_role()

    @property
    def key(self):
        return self.__stack.app.key

    @key.setter
    def key(self, key):
        self.__stack.app.key = key

    @property
    def aes_counter(self):
        return self.__stack.app.aes_counter

    @aes_counter.setter
    def aes_counter(self,aes_counter):
        self.__stack.app.aes_counter = aes_counter

    def start(self):
        self.__started = True
        self._enable_role()

    def stop(self):
        self.__started = False
        super().stop()

    @property
    def stack(self):
        return self.__stack

    @property
    def address(self):
        return self.__address

    @address.setter
    def address(self, address):
        self.__address = address
        self._enable_role()

    def wait_synchronization(self):
        return self.__stack.app.wait_synchronization()

    def wait_wakeup(self):
        return self.__stack.app.wait_wakeup()

    def on_pdu(self, packet):
        self.__stack.on_pdu(packet)

    def stream(self):
        for pdu in self.__stack.ll.data_stream():
            yield pdu



"""
WHAD device information module.

This module provides the `WhadDeviceInfo` class that stores all the information
about a compatible WHAD device.
"""
from typing import List, Optional
from packaging.version import parse

from whad.helpers import asciiz
from whad.hub import ProtocolHub

class DeviceInfo:
    """This class caches a device information related to its firmware, type,
    and supported domains and capabilities.

    :param DeviceInfoResp info_resp:  Whad message containing the device basic information.
    """

    @staticmethod
    def create(proto_ver: int = 2, max_speed: int = 115200, author: str = '', url: str = '',
               version: str = '0.0', dev_type: int = 0, dev_id: bytes = b'',
               capabilities: Optional[List[int]] = None) -> 'DeviceInfo':
        # Parse version string
        ver = parse(version)

        # Build an InfoQueryResp message
        msg = ProtocolHub().discovery.create_info_resp(
            type=dev_type, device_id=dev_id, proto_min_ver=proto_ver, max_speed=max_speed,
            fw_author=author.encode('utf-8'), fw_url=url.encode('utf-8'),
            fw_version_major=ver.major, fw_version_minor=ver.minor, fw_version_rev=ver.micro,
            capabilities=capabilities
        )
        return DeviceInfo(msg)

    def __init__(self, info_resp):
        """Populate device information object from WHAD's InfoQueryResp message.

        :param info_resp: WHAD InfoQueryResp message
        :type info_resp: whad.hub.discovery.InfoQueryResp
        """
        # Store device information
        self.__whad_version = info_resp.proto_min_ver
        self.__max_speed = info_resp.max_speed
        self.__fw_author = info_resp.fw_author
        self.__fw_url = info_resp.fw_url
        self.__fw_ver_maj = info_resp.fw_version_major
        self.__fw_ver_min = info_resp.fw_version_minor
        self.__fw_ver_rev = info_resp.fw_version_rev
        self.__device_type = info_resp.type
        self.__device_id = asciiz(info_resp.device_id)

        # Parse domains and capabilities
        self.__domains = {}
        self.__commands = {}
        for domain in info_resp.capabilities:
            self.__domains[domain & 0xFF000000] = domain & 0x00FFFFFF
            self.__commands[domain & 0xFF000000] = 0

    def add_supported_commands(self, domain, commands):
        """Adds supported command for a given domain to the device information.

        :param domain: target domain
        :param commands: bitmask representing the supported commands for the given domain
        """
        if domain in self.__domains:
            self.__commands[domain] = commands


    def has_domain(self, domain):
        """Determine if a domain is supported by this device.

        :param Domain domain: Domain to check.
        :returns: True if supported, False otherwise.
        """
        return domain in self.__domains


    def has_domain_cap(self, domain, capability):
        """Check if device supports a specific capability for a given domain.

        :param Domain domain: target domain
        :param Capability capability: capability to check
        """
        if domain in self.__domains:
            return self.__domains[domain] & (1 << capability) > 0
        return False


    def get_domain_capabilities(self, domain):
        """Get device domain capabilities.

        :param Domain domain: target domain
        :returns: Domain capabilities
        :rtype: Bitmask of capabilities
        """
        if domain in self.__domains:
            return self.__domains[domain]
        return None


    def get_domain_commands(self, domain):
        """Get supported commands for a specific domain

        :param Domain domain: Target domain
        :returns: Bitmask of supported commands
        """
        if domain in self.__commands:
            return self.__commands[domain]
        return None


    ##### Getters #####

    @property
    def version_str(self):
        """Returns the device firmware version string.

        :returns: Device firmware version string
        """
        return f'{self.__fw_ver_maj:d}.{self.__fw_ver_min:d}.{self.__fw_ver_rev:d}'

    @property
    def whad_version(self):
        """Returns the device supported whad version.
        """
        return self.__whad_version

    @property
    def fw_author(self) -> str:
        """Get the device firmware author name.

        :return: Device's firmware author name
        :return-type: str
        """
        return self.__fw_author.decode("utf-8")

    @property
    def fw_url(self) -> str:
        """Get the device firmware URL

        :return: Firmware related URL
        :return-type: str
        """
        return self.__fw_url.decode("utf-8")

    @property
    def max_speed(self) -> int:
        """Get maximum communication speed.

        :return: Maximum communication speed
        :return-type: int
        """
        return self.__max_speed

    @property
    def device_type(self):
        """Returns the device type.
        """
        return self.__device_type

    @property
    def device_id(self):
        """Returns the device id.
        """
        return self.__device_id

    @property
    def domains(self) -> list[int]:
        """Return the list of supported domains.
        """
        return list(self.__domains.keys())



"""
Pythonic Enhanced ShockBurst stack
"""
import logging
from typing import Optional

from whad.esb.stack.llm import LinkLayer
from whad.common.stack import Layer, alias, source

logger = logging.getLogger(__name__)

@alias('phy')
class ESBStack(Layer):
    """
    This class holds the main components of a (hackable) Enhanced ShockBurst stack:

    - the Link Layer (LL)
    - the Applicative Layer (APP)

    The Link Layer handles all the low-level operations, e.g., packet
    transmission, reception, synchronization and acknowledgements.
    Generally, a proprietary Applicative layer is implemented on the top of
    this stack (e.g., Unifying, Microsoft).
    """

    def __init__(self, connector, options: Optional[dict] = None):
        """
        Create an instance of Enhanced ShockBurst associated with a specific connector. This
        connector provides the transport layer.

        :param connector: Connector to use with this stack.
        :type connector: WhadDeviceConnector
        :param options: Options for this layer
        :type options: dict, optional
        """

        # Initialize with provided options or empty dict if options is None.
        super().__init__(options=options or {})

        # Save connector (used as PHY layer)
        self.__connector = connector

    @property
    def ll(self):
        """Return the associated link-layer instance

        :return: Link-layer instance
        :rtype: Layer
        """
        return self.get_layer('ll')


    @property
    def app(self):
        """Return the assiociated application layer

        :return: Application layer instance
        :rtype: Layer
        """
        return self.get_layer('app')

    def on_pdu(self, pdu):
        '''PDU callback.

        This callback handles a received PDU.
        '''
        logger.debug('received a PDU (%d bytes).', len(pdu))
        self.send('ll', pdu)


    @property
    def channel(self) -> int:
        """Return current channel number

        :return: Current channel number
        :rtype: int
        """
        return self.__connector.channel

    @channel.setter
    def channel(self, channel: int):
        """Set current channel number

        :param channel: New channel number to use
        :type channel: int
        """
        self.__connector.channel = channel

    @property
    def address(self) -> str:
        """Return current ESB address

        :return: Current connector address
        :rtype: str
        """
        return self.__connector.address

    @address.setter
    def address(self, address: str):
        """Set current ESB address

        :param address: New ESB address
        :type address: str
        """
        self.__connector.address = address

    @source('ll')
    def send_pdu(self, packet, channel: Optional[int] = None,
                 retransmission_count: Optional[int] = 1):
        """Send an ESB PDU

        :param packet: PDU to send
        :type packet: Packet
        :param channel: channel to use for transmission
        :type channel: int, optional
        :param retransmission_count: Maximum number of retransmission
        :type retransmission_count: int, optional
        """
        if channel is None:
            channel = self.__connector.channel
        logger.debug('transmitted a PDU (%d bytes).', len(packet))

        return self.__connector.send(packet, channel=channel,
                                     retransmission_count=retransmission_count)

ESBStack.add(LinkLayer)



"""ZigBee Application Sub Layer (NWK) exceptions
"""

class APSTimeoutException(Exception):
    """This exception is raised when a timeout occurs at APS level.
    """
    def __init__(self):
        super().__init__()



from whad.dot15d4.stack.mac.constants import MACAddressMode

class Cluster:
    """
    This class represents a ZigBee cluster.

    Every application exposes several input and output clusters, that will
    act as server / client for a specific task.
    """
    def __init__(self, cluster_id):
        # Cluster definition
        self.cluster_id = cluster_id

        # Reference to application is populated when the cluster is attached to an app.
        self.application = None

    def send_data(
                    self,
                    asdu,
                    destination_address_mode,
                    destination_address,
                    destination_endpoint,
                    alias_address=None,
                    alias_sequence_number=0,
                    radius=30,
                    security_enabled_transmission=False,
                    use_network_key=True,
                    acknowledged_transmission=False,
                    fragmentation_permitted=False,
                    include_extended_nonce=False
    ):
        """
        This method allows to send a Data PDU to the underlying application.
        """
        if self.application is None:
            return False

        return self.application.send_data(
            asdu,
            destination_address_mode,
            destination_address,
            destination_endpoint,
            alias_address=alias_address,
            alias_sequence_number=alias_sequence_number,
            radius=radius,
            security_enabled_transmission=security_enabled_transmission,
            use_network_key=use_network_key,
            acknowledged_transmission=acknowledged_transmission,
            fragmentation_permitted=fragmentation_permitted,
            include_extended_nonce=include_extended_nonce,
            cluster_id=self.cluster_id
        )

    def send_interpan_data(
                            self,
                            asdu,
                            asdu_handle=0,
                            source_address_mode=MACAddressMode.SHORT,
                            destination_pan_id=0xFFFF,
                            destination_address=0xFFFF,
                            destination_address_mode=MACAddressMode.SHORT,
                            acknowledged_transmission=False
    ):
        """
        This method allows to send an InterPAN PDU to the underlying application.
        """
        if self.application is None:
            return False

        return self.application.send_interpan_data(
            asdu,
            asdu_handle=asdu_handle,
            source_address_mode=source_address_mode,
            destination_pan_id=destination_pan_id,
            destination_address=destination_address,
            destination_address_mode=destination_address_mode,
            acknowledged_transmission=acknowledged_transmission,
            cluster_id=self.cluster_id
        )

    def on_data(
                self,
                asdu,
                source_address,
                source_address_mode,
                security_status,
                link_quality
    ):
        """
        This method is called when the underlying application forwards a Data PDU to the cluster.
        """
        pass

    def on_interpan_data(
                self,
                asdu,
                destination_pan_id,
                destination_address,
                source_pan_id,
                source_address,
                link_quality
    ):
        """
        This method is called when the underlying application forwards an InterPAN PDU to the cluster.
        """
        pass



import json
import csv
import sys
import os

def main():
    # Determine the input file path
    input_file = sys.argv[1] if len(sys.argv) > 1 else "./CLUES_data.json"
    output_file = "./bluetooth_uuids"

    # Check if the input file exists
    if not os.path.exists(input_file):
        print(f"Error: File '{input_file}' not found.")
        return

    try:
        # Open and load the JSON file
        with open(input_file, "r") as f:
            clues_data = json.load(f)

        # Open the CSV file for writing
        with open(output_file, "w", newline="") as csvfile:
            csv_writer = csv.writer(csvfile, quoting=csv.QUOTE_ALL)

            # Iterate through each entry in the JSON data
            for entry in clues_data:
                # Skip entries with "regex": true
                if "regex" in entry.keys():
                    continue

                # Extract the UUID
                first_field = entry.get("UUID")
                if not first_field:
                    continue  # Skip if UUID is missing

                # Determine the second CSV field
                company = entry.get("company", "Unknown")
                if "UUID_name" in entry:
                    second_field = f"{company}__{entry['UUID_name']}"
                else:
                    second_field = f"{company}__{entry['UUID_purpose']}"

                # Write the row to the CSV file
                csv_writer.writerow([first_field, second_field])

        print(f"CSV file '{output_file}' created successfully.")
        print(f"You must copy it to the correct location for Wireshark based on your OS.")
        print(f"E.g. on Linux and macOS: ~/.config/wireshark/bluetooth_uuids")
        print(f"E.g. on Windows: C:\Users\username\AppData\Roaming\Wireshark\bluetooth_uuids")

    except Exception as e:
        print(f"Error: {e}")

if __name__ == "__main__":
    main()
